"""
Property Collector helper module.
"""

import pyVmomi


# Shamelessly borrowed from:
# https://github.com/dnaeon/py-vconnector/blob/master/src/vconnector/core.py
def collect_properties(si, view_ref, obj_type, path_set=None,
                       include_mors=False):
    """
    Collect properties for managed objects from a view ref

    Check the vSphere API documentation for example on retrieving
    object properties:

        - http://goo.gl/erbFDz

    Args:
        si          (ServiceInstance): ServiceInstance connection
        view_ref (pyVmomi.vim.view.*): Starting point of inventory navigation
        obj_type      (pyVmomi.vim.*): Type of managed object
        path_set               (list): List of properties to retrieve
        include_mors           (bool): If True include the managed objects
                                       refs in the result

    Returns:
        A list of properties for the managed objects

    """
    collector = si.content.propertyCollector

    # Create object specification to define the starting point of
    # inventory navigation
    obj_spec = pyVmomi.vmodl.query.PropertyCollector.ObjectSpec()
    obj_spec.obj = view_ref
    obj_spec.skip = True

    # Create a traversal specification to identify the path for collection
    traversal_spec = pyVmomi.vmodl.query.PropertyCollector.TraversalSpec()
    traversal_spec.name = 'traverseEntities'
    traversal_spec.path = 'view'
    traversal_spec.skip = False
    traversal_spec.type = view_ref.__class__
    obj_spec.selectSet = [traversal_spec]

    # Identify the properties to the retrieved
    property_spec = pyVmomi.vmodl.query.PropertyCollector.PropertySpec()
    property_spec.type = obj_type

    if not path_set:
        property_spec.all = True

    property_spec.pathSet = path_set

    # Add the object and property specification to the
    # property filter specification
    filter_spec = pyVmomi.vmodl.query.PropertyCollector.FilterSpec()
    filter_spec.objectSet = [obj_spec]
    filter_spec.propSet = [property_spec]

    # Retrieve properties
    props = collector.RetrieveContents([filter_spec])

    data = []
    for obj in props:
        properties = {}
        for prop in obj.propSet:
            properties[prop.name] = prop.val

        if include_mors:
            properties['obj'] = obj.obj

        data.append(properties)
    return data


def get_container_view(si, obj_type, container=None):
    """
    Get a vSphere Container View reference to all objects of type 'obj_type'

    It is up to the caller to take care of destroying the View when no longer
    needed.

    Args:
        obj_type (list): A list of managed object types

    Returns:
        A container view ref to the discovered managed objects
    """
    if not container:
        container = si.content.rootFolder

    view_ref = si.content.viewManager.CreateContainerView(
        container=container,
        type=obj_type,
        recursive=True
    )
    return view_ref


def search_for_obj(content, vim_type, name, folder=None, recurse=True):
    """
    Search the managed object for the name and type specified

    Sample Usage:

    get_obj(content, [vim.Datastore], "Datastore Name")
    """
    if folder is None:
        folder = content.rootFolder

    obj = None
    container = content.viewManager.CreateContainerView(folder, vim_type, recurse)

    for managed_object_ref in container.view:
        if managed_object_ref.name == name:
            obj = managed_object_ref
            break
    container.Destroy()
    return obj


def get_all_obj(content, vim_type, folder=None, recurse=True):
    """
    Search the managed object for the name and type specified

    Sample Usage:

    get_obj(content, [vim.Datastore])
    """
    if not folder:
        folder = content.rootFolder

    obj = {}
    container = content.viewManager.CreateContainerView(folder, vim_type, recurse)

    for managed_object_ref in container.view:
        obj[managed_object_ref] = managed_object_ref.name

    container.Destroy()
    return obj


def get_obj(content, vim_type, name, folder=None, recurse=True):
    """
    Retrieves the managed object for the name and type specified
    Throws an exception if of not found.

    Sample Usage:

    get_obj(content, [vim.Datastore], "Datastore Name")
    """
    obj = search_for_obj(content, vim_type, name, folder, recurse)
    if not obj:
        raise RuntimeError("Managed Object " + name + " not found.")
    return obj



# import subprocess
# from libvirt import *
# from xml.dom import minidom


# def connect(host: str, username: str):
#     conn = open(f"qemu+ssh://{username}@{host}/system")
#     if not conn:
#         raise SystemExit("Failed to open connection to qemu:///system")

#     # caps = conn.getCapabilities()  # caps will be a string of XML
#     # print("Capabilities:\n" + caps)
#     return conn


# def vm(conn: virConnect, id):
#     dom: virDomain = conn.lookupByID(id)
#     if not dom:
#         raise SystemExit("Failed to find domain ID 5")

#     raw_xml = dom.XMLDesc()
#     xml = minidom.parseString(raw_xml)
#     domainTypes = xml.getElementsByTagName("type")
#     for domainType in domainTypes:
#         print(domainType.getAttribute("machine"))
#         print(domainType.getAttribute("arch"))


# def disks(dom: virDomain):
#     raw_xml = dom.XMLDesc()

#     xml = minidom.parseString(raw_xml)

#     diskTypes = xml.getElementsByTagName("disk")
#     for diskType in diskTypes:
#         print(
#             "disk: type="
#             + diskType.getAttribute("type")
#             + " device="
#             + diskType.getAttribute("device")
#         )
#         diskNodes = diskType.childNodes
#         for diskNode in diskNodes:
#             if diskNode.nodeName[0:1] != "#":
#                 print("  " + diskNode.nodeName)
#                 for attr in diskNode.attributes.keys():
#                     print(
#                         "    "
#                         + diskNode.attributes[attr].name
#                         + " = "
#                         + diskNode.attributes[attr].value
#                     )


# def networks(dom):
#     raw_xml = dom.XMLDesc()
#     xml = minidom.parseString(raw_xml)

#     interfaceTypes = xml.getElementsByTagName("interface")
#     for interfaceType in interfaceTypes:
#         print("interface: type=" + interfaceType.getAttribute("type"))
#         interfaceNodes = interfaceType.childNodes
#         for interfaceNode in interfaceNodes:
#             if interfaceNode.nodeName[0:1] != "#":
#                 print("  " + interfaceNode.nodeName)
#                 for attr in interfaceNode.attributes.keys():
#                     print(
#                         "    "
#                         + interfaceNode.attributes[attr].name
#                         + " = "
#                         + interfaceNode.attributes[attr].value
#                     )


# def pools(conn: virConnect):
#     pools = conn.listAllStoragePools()
#     if not pools:
#         return list()
#         # raise SystemExit("Failed to locate any StoragePool objects.")

#     return [pool.name() for pool in pools if pool.isActive() == True]


# def newdisk(conn:virConnect):
#     stgvol_xml = """
#     <volume>
#     <name>sparse.img</name>
#     <allocation>0</allocation>
#     <capacity unit="G">2</capacity>
#     <target>
#         <path>/var/lib/libvirt/images/sparse.img</path>
#         <permissions>
#         <owner>107</owner>
#         <group>107</group>
#         <mode>0744</mode>
#         <label>virt_image_t</label>
#         </permissions>
#     </target>
#     </volume>"""

#     pool = "default"

#     pool = conn.storagePoolLookupByName(pool)
#     if not pool:
#         raise SystemExit("Failed to locate any StoragePool objects.")

#     stgvol = pool.createXML(stgvol_xml, 0)
#     if not stgvol:
#         raise SystemExit("Failed to create a  StorageVol objects.")

#     # remove the storage volume
#     # physically remove the storage volume from the underlying disk media
#     stgvol.wipe()
#     # logically remove the storage volume from the storage pool
#     stgvol.delete()


# def clonedisk(conn:virConnect):
#     stgvol_xml = """
#     <volume>
#         <name>sparse.img</name>
#         <allocation>0</allocation>
#         <capacity unit="G">2</capacity>
#         <target>
#             <path>/var/lib/libvirt/images/sparse.img</path>
#             <permissions>
#                 <owner>107</owner>
#                 <group>107</group>
#                 <mode>0744</mode>
#                 <label>virt_image_t</label>
#             </permissions>
#         </target>
#     </volume>"""

#     stgvol_xml2 = """
#     <volume>
#         <name>sparse2.img</name>
#         <allocation>0</allocation>
#         <capacity unit="G">2</capacity>
#         <target>
#             <path>/var/lib/libvirt/images/sparse.img</path>
#             <permissions>
#                 <owner>107</owner>
#                 <group>107</group>
#                 <mode>0744</mode>
#                 <label>virt_image_t</label>
#             </permissions>
#         </target>
#     </volume>"""

#     pool = conn.storagePoolLookupByName(pool)

#     if not pool:
#         raise SystemExit("Failed to locate any StoragePool objects.")

#     # create a new storage volume
#     stgvol = pool.createXML(stgvol_xml, 0)
#     if not stgvol:
#         raise SystemExit("Failed to create a  StorageVol object.")

#     # now clone the existing storage volume
#     print("This could take some time...")
#     stgvol2 = pool.createXMLFrom(stgvol_xml2, stgvol, 0)
#     if not stgvol2:
#         raise SystemExit("Failed to clone a  StorageVol object.")

#     stgvol2.wipe()
#     stgvol2.delete()

#     stgvol.wipe()
#     stgvol.delete()


# def newvm(conn:virConnect):
#     xmlconfig = """
# <domain type="kvm">
#   <name>demo</name>
#   <uuid>c7a5fdbd-cdaf-9455-926a-d65c16db1809</uuid>
#   <memory>500000</memory>
#   <vcpu>1</vcpu>

#   <os>
#     <type arch="x86_64" machine="pc">hvm</type>
#     <boot dev="hd"/>
#     <boot dev="cdrom"/>
#   </os>

#   <clock offset="utc"/>
#   <on_poweroff>destroy</on_poweroff>
#   <on_reboot>restart</on_reboot>
#   <on_crash>destroy</on_crash>
#   <devices>
#     <emulator>/usr/bin/qemu-kvm</emulator>
#     <disk type="file" device="cdrom">
#       <source file="/var/lib/libvirt/images/rhel5-x86_64-dvd.iso"/>
#       <target dev="hdc" bus="ide"/>
#     </disk>

#     <disk type="file" device="disk">
#       <source file="/var/lib/libvirt/images/demo.img"/>
#       <driver name="qemu" type="raw"/>
#       <target dev="hda"/>
#     </disk>
#     <interface type="bridge">
#       <mac address="52:54:00:d8:65:c9"/>
#       <source bridge="br0"/>
#     </interface>
#     <input type="mouse" bus="ps2"/>
#     <graphics type="vnc" port="-1" listen="127.0.0.1"/>
#   </devices>
# </domain>
# """
#     dom = conn.defineXML(xmlconfig)

#     if not dom:
#         raise SystemExit("Failed to define a domain from an XML definition")

#     if dom.create(dom) < 0:
#         raise SystemExit("Can not boot guest domain")

#     dom.setAutostart(1)  # enable autostart

#     print("Guest " + dom.name() + " has booted")

# def qemu_img_create(file_path, fmt="qcow2", size=None, backing_file=None):
#     cmd = ["qemu-img", "create", "-f", fmt]
#     if backing_file:
#         cmd.extend(["-b", backing_file])
#     cmd.append(file_path)
#     if size:
#         cmd.append(size)
#     subprocess.check_call(cmd)
    
# def build_domain(conn: virConnect, base_img, top_layer):
#     xmlconfig = """
# <domain type="kvm">
#   <name>demo</name>
#   <uuid>c7a5fdbd-cdaf-9455-926a-d65c16db1809</uuid>
#   <memory>500000</memory>
#   <vcpu>1</vcpu>

#   <os>
#     <type arch="x86_64" machine="pc">hvm</type>
#     <boot dev="hd"/>
#     <boot dev="cdrom"/>
#   </os>

#   <clock offset="utc"/>
#   <on_poweroff>destroy</on_poweroff>
#   <on_reboot>restart</on_reboot>
#   <on_crash>destroy</on_crash>
#   <devices>
#     <emulator>/usr/bin/qemu-kvm</emulator>
#     <disk type="file" device="cdrom">
#       <source file="/var/lib/libvirt/images/rhel5-x86_64-dvd.iso"/>
#       <target dev="hdc" bus="ide"/>
#     </disk>

#     <disk type="file" device="disk">
#       <source file="/var/lib/libvirt/images/demo.img"/>
#       <driver name="qemu" type="raw"/>
#       <target dev="hda"/>
#     </disk>
#     <interface type="bridge">
#       <mac address="52:54:00:d8:65:c9"/>
#       <source bridge="br0"/>
#     </interface>
#     <input type="mouse" bus="ps2"/>
#     <graphics type="vnc" port="-1" listen="127.0.0.1"/>
#   </devices>
# </domain>
# """

#     qemu_img_create(base_img, size="100M")
#     qemu_img_create(top_layer, backing_file=base_img)
#     dom = conn.createXML(xmlconfig.format(disk_img=top_layer), 0)
#     return dom



# # conn = connect("10.1.1.5", "ezmeral")

# # pool = pools(conn).pop()

# # print(pool)

# # conn.close()




#!/usr/bin/env python3
"""
Copyright 2023 Erdinc Kaya
DESCRIPTION:
    Deploy VMs on my lab and configure them for Ezmeral
USAGE EXAMPLE:
    > ezlab
"""

import json
import logging
import secrets
# import sys
from nicegui import ui, app, native
# sys.stdout = open("ezlogs.txt", "w")  # divert stdout to logs.txt file

from ezlab.app import LogElementHandler, ezmeral_menu, infra_menu, settings_menu

from ezlab.parameters import DF, SUPPORTED_HVES, UA

logger = logging.getLogger("ezlab")

# Initialize
# Top-level keys in config
for key in ["target", "ezui", "config", UA, DF, *SUPPORTED_HVES]:
    if not key in app.storage.general.keys():
        app.storage.general[key] = {}

app.storage.general["target"]["connected"] = False
app.native.start_args["debug"] = True

# Index page
@ui.page("/")
def home():
    # initial status
    app.storage.user["busy"] = False

    # Config show dialog
    with ui.dialog() as config_show, ui.card().classes("w-full h-full"):
        config_json = json.dumps(app.storage.general, indent=2)
        code_text = ui.code(
            config_json, language="json"
        ).classes("w-full text-wrap")
        with ui.row():
            ui.button(
                "Save (useless on app mode)",
                icon="save",
                on_click=lambda: ui.download(config_json.encode(), "ezlab-settings.json"),
            )
            ui.button(
                "Close",
                icon="cancel",
                on_click=config_show.close,
            )
    def save_config(val: str, dialog):
        try:
            for key, value in json.loads(val.replace("\n", "")).items():
                app.storage.general[key] = value
            for menu in (settings_menu, ezmeral_menu, infra_menu):
                menu.refresh()
            dialog.close()
            ui.notify("Settings loaded", type="positive")
        except (TypeError, json.decoder.JSONDecodeError, ValueError) as error:
            ui.notify("Not a valid json", type="negative")
            print(error)

    # Config load dialog
    with ui.dialog() as config_set, ui.card().classes("w-full h-full"):
        jsontext = (
            ui.textarea()
            .props("stack-label=json autogrow filled")
            .classes("h-dvh w-full text-wrap")
        )
        ui.button("Save", on_click=lambda _: save_config(jsontext.value, config_set))

    # Header
    with ui.header(elevated=True).classes("items-center justify-between") as header:
        # home button
        ui.label("Ezlab").classes("text-bold text-lg")

        ui.space()

        with ui.row().classes("items-center"):
            ui.label("Settings")
            ui.button(icon="download", on_click=config_show.open)
            ui.button(icon="upload", on_click=config_set.open)

    # Footer
    with ui.footer() as footer:
        with ui.row().classes("w-full"):
            ui.label("Log")
            ui.space()
            ui.spinner("ios", size="2em", color="red").bind_visibility_from(
                app.storage.user, "busy"
            )
            ui.icon("check_circle", size="2em", color="green").bind_visibility_from(
                app.storage.user, "busy", lambda x: not x
            )
        log = ui.log().classes("w-full h-48 text-teal-400")
        logger.addHandler(LogElementHandler(log))

    # Content

    # Settings
    settings_menu()

    ui.separator()

    # Infrastructure
    infra_menu()

    ui.separator()

    # Ezmeral Products
    ezmeral_menu()

ezmeral_icon = """
<svg width="48" height="24" viewBox="0 0 48 24" fill="none" xmlns="http://www.w3.org/2000/svg">
    <path d="M7 8H41V16H7V8Z" fill="#01A982"/>
    <path d="M1 8H7V16H1V8Z" fill="#00775B"/>
    <path d="M41 8H47V16H41V8Z" fill="#00775B"/>
    <path d="M7 16H41V22H7V16Z" fill="#00775B"/>
    <path d="M7 2H41V8H7V2Z" fill="#00C781"/>
    <path d="M1 8L7 2V8H1Z" fill="#01A982"/>
    <path d="M1 16L7 22V16H1Z" fill="#01A982"/>
    <path d="M47 8L41 2V8H47Z" fill="#01A982"/>
    <path d="M47 16L41 22V16H47Z" fill="#01A982"/>
</svg>
"""

def enter():
    # ui.run(
    #     title="Ezlab",
    #     dark=None,
    #     favicon=ezmeral_icon,
    #     storage_secret=secrets.token_urlsafe(),
    #     show=True,
    #     # reload=False,
    # )
    ui.run(
        title="Ezlab",
        dark=None,
        native=True,
        window_size=(1024, 1024),
        # frameless=True,
        storage_secret=secrets.token_urlsafe(),
        reload=False,
        port=native.find_open_port(),
    )



import os
import platform
from socket import gethostbyname
from typing import Annotated, List
import typer
from kubernetes import client as k8s_client, config as k8s_config
import docker
import yaml

from .common import (
    execute,
    fail,
    find_app,
    get_proxy_environment,
    toB64,
    validate_ip,
    vm_proxy,
)
from . import config as ezconfig
from .parameters import ez_product, ez_installers

app = typer.Typer(add_completion=True, no_args_is_help=True)

defaults = ezconfig.get()

ez_cluster_name = "ezlab"

ezfab_yaml = os.path.abspath("ezfab.yaml")
orchestratorkubeconfig = os.path.abspath("ezfab-orchestrator-kubeconfig")
workloadkubeconfig = os.path.abspath("ezfab-workload-kubeconfig")
workloadcr = os.path.abspath("ezkf-workload-deploy-cr.yaml")
ezfabricctl = find_app("ezfabricctl")
ezfabrelease = os.path.abspath("ezfab-release.tgz")


def init_orchestrator(
    ip: Annotated[str, typer.Option("--ip", "-i")],
    ezfabrelease: Annotated[str, typer.Option("--ezfab-release", "-e")] = ezfabrelease,
):
    """Install UA orchestrator

    Args:
        ip (str): orchestrator IP address
        ezfabrelease (str): full path to ezfab-release.tgz file
    """

    write_ezfabyaml(filename=ezfab_yaml, purpose="coordinator_init", orchestrator_ip=ip)

    if ezfabricctl is None:
        fail(
            f"ezfabricctl not found, check if you get ezfabricctl_{platform.system().lower()}_amd64 binary from the installer container and marked it as executable."
        )

    if os.path.isfile(os.path.abspath(ezfabrelease)):
        print("Starting coordinator deployment...")
        for out in execute(
            f"{ezfabricctl} o init -p {ezfabrelease} -i {ezfab_yaml} --save-kubeconfig {orchestratorkubeconfig}"
        ):
            print(out)
            # if "Save kubeconfig" in out and orchestratorkubeconfig in out:
            #     shutil.copyfile(
            #         orchestratorkubeconfig,
            #         os.path.expanduser("$HOME/.kube/ezfab-orchestrator"),
            #     )
    else:
        fail("Provide full path to ezfab-release.tgz file")


def add_hosts_to_pool(
    worker_ips: Annotated[
        List[str], typer.Option("--worker_ip", "-w", help="One or more")
    ],
    kubeconf: Annotated[
        str, typer.Option("--kubeconfig", "-c")
    ] = orchestratorkubeconfig,
):
    """Add hosts to UA pool

    Args:
        worker_ips (List(str)): hostname/IP address to add (can be provided multiple times)
        kubeconf (str): path to orchestrator kubeconfig
    """

    if ezfabricctl is None:
        fail(
            f"ezfabricctl not found, check if you get ezfabricctl_{platform.system().lower()}_amd64 binary from the installer container and marked it as executable."
        )

    write_ezfabyaml(filename=ezfab_yaml, purpose="poolhost_init", workers=worker_ips)

    for out in execute(
        f"{ezfabricctl} ph init -i {ezfab_yaml} -c {os.path.abspath(kubeconf)}"
    ):
        print(out)


@app.command(no_args_is_help=True)
def create_workload_cluster(
    orchestrator_ip: Annotated[str, typer.Option("--orchestrator-ip", "-o")],
    kubeconf: Annotated[
        str, typer.Option("--kubeconfig", "-c")
    ] = orchestratorkubeconfig,
    cluster_name: Annotated[str, typer.Option("--clusterName", "-n")] = ez_cluster_name,
):
    """Create UA workload cluster

    Args:
        orchestrator (str): controller IP address
        kubeconf (str): path to orchestrator kubeconfig
    """

    if ezfabricctl is None:
        fail(
            f"ezfabricctl not found, check if you get ezfabricctl_{platform.system().lower()}_amd64 binary from the installer container and marked it as executable."
        )

    # use full path for the config file
    kubeconf = os.path.abspath(kubeconf)

    write_ezfabyaml(
        filename=ezfab_yaml,
        purpose="workload_init",
        orchestrator_ip=orchestrator_ip,
    )

    for out in execute(f"{ezfabricctl} workload init -i {ezfab_yaml} -c {kubeconf}"):
        print(out)

    # If the orchestrator is ready
    if os.path.isfile(kubeconf):
        # Save the workload kubeconfig
        for out in execute(
            f"{ezfabricctl} workload get kubeconfig -n {cluster_name} -c {kubeconf} --save-kubeconfig {workloadkubeconfig}"
        ):
            print(out)
        print(f"Workload kubeconfig saved as {workloadkubeconfig}")
    else:
        fail(f"Orchestrator kubeconfig file {kubeconf} not found!")


@app.command(no_args_is_help=True)
def applyCr(
    cluster_name: Annotated[str, typer.Option("--clusterName", "-n")] = ez_cluster_name,
):
    # Deploy the workload CR
    k8s_config.load_kube_config(config_file=orchestratorkubeconfig)

    # with open(workloadkubeconfig, "r") as f:
    #     kubeconfig = toB64(f.read())

    # k8s_config = {
    #     "apiVersion": "v1",
    #     "kind": "Secret",
    #     "metadata": {
    #         "name": "k8sconfig",
    #         "namespace": cluster_name
    #     },
    #     "data": {
    #         "kubeconfig": kubeconfig
    #     }
    # }

    v1 = k8s_client.CoreV1Api()
    secrets = v1.list_namespaced_secret(cluster_name)
    if "authconfig" not in [secret.metadata.name for secret in secrets.items]:
        admin = {
            "admin_user": {
                "username": defaults["CLOUDINIT"]["vm_username"],
                "fullname": "Ezlab Admin User",
                "email": f"{defaults['CLOUDINIT']['vm_username']}@{defaults['NETWORK']['vm_domain']}",
                "password": defaults["CLOUDINIT"]["vm_password"],
            }
        }
        auth_config = {
            "apiVersion": "v1",
            "kind": "Secret",
            "metadata": {"name": "authconfig", "namespace": cluster_name},
            "data": {"internal_auth": toB64(str(admin))},
        }
        v1.create_namespaced_secret(cluster_name, body=auth_config)

    api = k8s_client.CustomObjectsApi()

    if cluster_name not in [
        deployment["metadata"]["name"]
        for deployment in getDeploymentCRs(api, cluster_name)
    ]:
        cr = api.create_namespaced_custom_object(
            namespace=cluster_name,
            group="ezkfops.hpe.ezkf-ops.com",
            version="v1alpha1",
            plural="ezkfworkloaddeploys",
            body=get_ezkfcr_json(cluster_name=cluster_name),
        )
        if cluster_name in cr["metadata"]["name"]:
            print(f"Deployment created for {cluster_name}")
    else:
        print(f"{cluster_name} exists, remove from EzkfWorkloadDeploy CR to retry")

    fail(
        "This process doesn't work with the error: ERROR: in cluster_deploy - Error creating serializer: {'deployname': [ErrorDetail(string='This field may not be null.', code='null')], 'domainname': [ErrorDetail(string='This field may not be null.', code='null')], 'deployallapps': [ErrorDetail(string='This field may not be null.', code='null')], 'deployallinfra': [ErrorDeta..."
    )


@app.command(no_args_is_help=True)
def remove(
    host: Annotated[str, typer.Option("--host", "-h")],
    ezfabrelease: Annotated[str, typer.Option("--ezfab-release", "-e")] = ezfabrelease,
    kubeconf: Annotated[
        str, typer.Option("--kubeconfig", "-c")
    ] = orchestratorkubeconfig,
):
    """Remove UA

    Args:
        host (str): hostname/IP address to destroy
    """

    if ezfabricctl is None:
        fail(
            f"ezfabricctl not found, check if you get ezfabricctl_{platform.system().lower()}_amd64 binary from the installer container and marked it as executable."
        )

    write_ezfabyaml(
        filename=ezfab_yaml, purpose="coordinator_init", orchestrator_ip=host
    )

    if os.path.isfile(os.path.abspath(ezfabrelease)):
        print("Destroying EZUA...")
        for out in execute(
            f"{ezfabricctl} o destroy --force -p {ezfabrelease} -i {ezfab_yaml} -c {os.path.abspath(kubeconf)}"
        ):
            print(out)
        print("Removing kubeconfig and cr files")
        for file in [orchestratorkubeconfig, workloadkubeconfig, ezfab_yaml]:
            if os.path.isfile(file):
                os.remove(file)
            else:
                print(f"{os.path.basename(file)} not found, skipping...")

    else:
        fail("Provide full path to ezfab-release.tgz file")


def write_ezfabyaml(
    filename: str,
    purpose: str,
    orchestrator_ip: str = "",
    workers: list = [],
    cluster_name: str = ez_cluster_name,
) -> bool:
    """Generate yaml as input for ezfabctl
    TODO: should be using yaml module for proper yaml generation
    """

    response = False

    try:
        # Just in case hostname/fqdn is provided for orchestrator (we assume IP address for the rest)
        if not validate_ip(orchestrator_ip):
            try:
                orchestrator_ip = gethostbyname(orchestrator_ip)
            except Exception as error:
                fail(f"Cannot get host IP: {error}")

        # get private key from public key path, assuming private key file has the same name with public key without the .pub suffix
        ssh_prvkey = ""
        with open(
            os.path.abspath(
                os.path.expanduser(
                    defaults["CLOUDINIT"]["ssh_keyfile"].replace(".pub", "")
                )
            ),
            "r",
        ) as f:
            ssh_prvkey = f.read()

        obj = {
            "defaultHostCredentials": {
                "sshUserName": defaults["CLOUDINIT"]["vm_username"],
                "sshPrivateKey": toB64(ssh_prvkey),
            },
            "airgap": {
                "registryUrl": "lr1-bd-harbor-registry.mip.storage.hpecorp.net/develop/",
                "registryInsecure": True,
            },
            "hosts": [],
        }

        if purpose == "coordinator_init":
            obj["hosts"] = [{"host": orchestrator_ip}]

        elif purpose == "poolhost_init":
            for host in workers:
                obj["hosts"].append(
                    {
                        "host": host,
                        "labels": {
                            "role": "worker"
                            if workers.index(host) > 0
                            # first one is the control plane (K8s master)
                            else "controlplane"
                        },
                    }
                )

        elif purpose == "workload_init":
            obj["workload"] = {
                "deployEnv": "ezkube",
                "workloadType": "ezua",
                "clusterName": cluster_name,
                "resources": {"vcpu": 96},
                "controlplane": {"controlPlaneEndpoint": workers[0]},
                "controlPlaneHostLabels": {"role": "controlplane"},
                "workerHostLabels": {"role": "worker"},
            }

        with open(filename, "w") as f:
            f.write(yaml.safe_dump(obj))

    except Exception as error:
        print(error)
        response = False

    return response


def getDeploymentCRs(
    api: k8s_client.CustomObjectsApi, namespace: str = ez_cluster_name
):
    crs = api.list_namespaced_custom_object(
        namespace=namespace,
        group="ezkfops.hpe.ezkf-ops.com",
        version="v1alpha1",
        plural="ezkfworkloaddeploys",
    )
    if crs is not None and len(crs) > 0:
        _, deployments, _, _ = crs.values()
    else:
        deployments = []

    return deployments


def get_ezkfcr_json(cluster_name: str = ez_cluster_name):
    """Generate yaml for EzkfWorkloadDeploy CR"""

    cr = {
        "apiVersion": "ezkfops.hpe.ezkf-ops.com/v1alpha1",
        "kind": "EzkfWorkloadDeploy",
        "metadata": {"name": cluster_name, "namespace": cluster_name},
        "spec": {
            "clustername": cluster_name,
            "deployallapps": True,
            "deployallinfra": True,
            "deployenv": "ezkube",
            "deploytarget": "pph",
            "domainname": "testdom.com",
            "workloadtype": "ezua",
            "authconfig": {"secret_name": "authconfig"},
        },
    }
    # "k8sconfig": { "secret_name": "k8sconfig" },
    # "tlsconfig":{ "secret_name": "tlsconfig" },
    if vm_proxy != "":
        cr["spec"]["proxy"] = {
            "httpProxy": vm_proxy,
            "httpsProxy": vm_proxy,
            "noProxy": get_proxy_environment().split("no_proxy=")[1],
        }

    # # TODO: Not tested
    # if "airgap_registry" in defaults["NETWORK"]:
    #     cr["spec"]["airgap"] = {
    #         "secret_name": "airgap",
    #         "registryUrl": {defaults["NETWORK"]["airgap_registry"]},
    #         "registryInsecure": {
    #             "true"
    #             if defaults["NETWORK"]["airgap_registry"].split("://")[0] == "http"
    #             else "false"
    #         },
    #     }

    return cr


@app.command(no_args_is_help=True)
def with_rest(cluster_name: str = ez_cluster_name):
    image = ez_installers[ez_product.ua.value]["image"]
    img = typer.prompt(f"EZUA UI image: ", default=image)

    client = docker.from_env()

    if client.containers.list(filters={"ancestor": img}):
        print("Container already running...")
    else:
        env = (
            {
                "HTTP_PROXY": vm_proxy,
                "HTTPS_PROXY": vm_proxy,
                "NO_PROXY": get_proxy_environment().split("no_proxy=")[1].strip(),
            }
            if vm_proxy != ""
            else {}
        )
        # print(f"Using environment vars for: {env}")
        try:
            ctr = client.containers.run(
                image=img,
                detach=True,
                privileged=True,
                restart_policy={"Name": "always"},
                environment=env,
                name="ezua-installer",
                ports={"8080/tcp": 8080},
                tty=True,
            )
            print(ctr)
        # except docker.errors.ImageNotFound:
        #     fail("Image not found!")
        # except docker.errors.APIError as error:
        #     fail(f"Server returned error: {error}")
        except Exception as error:
            fail(error)

    # typer.launch(ez_installers[ez_product.ua.value]["url"])

    # Create ezfab yaml and submit to UI
    url = f"{ez_installers[ez_product.ua.value].url}/api/v1/af/"
    installer_info_json = {
        "type": "new",
        "deployname": cluster_name,
        "deploytarget": "pph",
        "provider": "vm",
        "infraconfig": {"high_availability": False, "vcpu": 96},
        "domainname": defaults["NETWORK"]["vm_domain"],
    }

    payload_json = {
        "deployname": cluster_name,
        "domainname": defaults["NETWORK"]["vm_domain"],
        "deployallapps": True,
        "deployallinfra": True,
        "infraconfig": {"vcpu": 96, "high_availability": False},
    }


# Will be available on FY24-Q1 release
def precheck():
    """Fabric pre-checks for roles"""

    payload = """
defaultHostCredentials:
  sshUserName: "user"
  sshPassword: "base64encode password"
prechecks:
  coordinator:
    controlplane:
    - host: "IP address or FQDN"
      #hostCredentials: {} # per-host credentials, same format as defaultHostCredentials
      #hostSudoPrefix: "" # per-host sudo prefix
  ua:
    controlplane:
    - host: "IP address or FQDN"
    worker:
    - host: "IP address or FQDN"
    - host: "IP address or FQDN"
"""


@app.command(no_args_is_help=True)
def install(
    orchestrator_ip: Annotated[str, typer.Option("--orchestrator-ip", "-o")],
    workers: Annotated[
        List[str], typer.Option("--worker_ip", "-w", help="One or more")
    ],
):
    init_orchestrator(ip=orchestrator_ip)
    add_hosts_to_pool(worker_ips=workers)
    create_workload_cluster(orchestrator_ip=orchestrator_ip)



#!/usr/bin/env python3

# Dependent packages
import asyncio
import logging
import os
import random
import string
import time
from nicegui import app, ui, run
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from queue import Empty, Queue


# Module packages
from ezlab.parameters import *
from ezlab.utils import *
from ezlab.ezmeral.ezdf import *
# from .ezua import *
# from . import kvm
from ezlab.infra import pve
# from . import vmw


class ConnectionContext:
    connection = None

    def __init__(self) -> None:
        pass

# Module Globals
logger = logging.getLogger("ezlab")
logger.setLevel(level=logging.DEBUG)
ez = ConnectionContext()

# Utility
class LogElementHandler(logging.Handler):
    """A logging handler that emits messages to a log element."""

    def __init__(self, element: ui.log, level: int = logging.NOTSET) -> None:
        self.element = element
        super().__init__(level)

    def emit(self, record: logging.LogRecord) -> None:
        try:
            msg = self.format(record)
            self.element.push(msg)
            with open("ezrun.log", "a") as f:
                print(msg, file=f)
        except Exception:
            self.handleError(record)


@ui.refreshable
def ssh_keys_ui():
    if is_file(SSHKEYNAME):
        ui.label(f"Existing Key: {os.path.abspath(SSHKEYNAME)}")
    with ui.row():
        ui.upload(
            label="Import",
            on_upload=lambda e: save_sshkey(e.content.read().decode("utf-8")),
            on_rejected=lambda x: ui.notify("No file selected"),
            max_file_size=1_000_000,
            max_files=1,
            auto_upload=True,
        ).props("accept=*")

        ui.button("Create New", on_click=create_sshkey)


async def confirm_overwrite(file):
    with ui.dialog() as dialog, ui.card():
        ui.label(f"{file} will be overwritten, are you sure?")
        with ui.row():
            ui.button("Yes", on_click=lambda: dialog.submit("Yes"))
            ui.button("No", on_click=lambda: dialog.submit("No"))
    return await dialog


async def save_sshkey(key: str):
    # verify format and get private key
    private_key = get_privatekey_from_string(key)
    if private_key is None:
        ui.notify("Not a known key format", type="negative")
        return None

    if is_file(SSHKEYNAME):
        if await confirm_overwrite(SSHKEYNAME) != "Yes":
            ui.notify("Cancelled", type="warning")
            return None

    private_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.OpenSSH,
        encryption_algorithm=serialization.NoEncryption(),
    )
    with open(SSHKEYNAME, "wb") as private_keyfile:
        private_keyfile.write(private_pem)

    public_pem = (
        private_key.public_key()
        .public_bytes(
            encoding=serialization.Encoding.OpenSSH,
            format=serialization.PublicFormat.OpenSSH,
        )
        .decode("utf-8")
    )

    with open(f"{SSHKEYNAME}.pub", "w") as public_keyfile:
        public_keyfile.write(public_pem)

    app.storage.general["config"]["privatekey"] = key
    app.storage.general["config"]["privatekeyfile"] = os.path.abspath(SSHKEYNAME)

    ui.notify(f"Key saved as {os.path.abspath(SSHKEYNAME)}", type="positive")
    ssh_keys_ui.refresh()
    return True


async def create_sshkey():
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    private_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.OpenSSH,
        encryption_algorithm=serialization.NoEncryption(),
    ).decode("utf-8")
    await save_sshkey(private_pem)


# Top-level UI menus
@ui.refreshable
def infra_menu():
    with (
        ui.expansion("VMs", icon="computer", caption="create and configure")
        .classes("w-full")
        .classes("text-bold") as virtualmachines
    ):
        with ui.row().classes("w-full items-center justify-between"):
            # target connection
            connect_ui()
            # new vm button
            ui.label().bind_text_from(
                app.storage.general["target"],
                "hve",
                backward=lambda x: f"{x} connected",
            ).bind_visibility_from(app.storage.general["target"], "connected")
            ui.button(
                "New VM", on_click=new_vm
            ).bind_visibility_from(app.storage.general["target"], "connected").bind_enabled_from(app.storage.user, "busy", backward=lambda x: not x)

    virtualmachines.bind_value(app.storage.general["ezui"], "virtualmachines")


@ui.refreshable
def ezmeral_menu():
    with (
        ui.expansion(
            "Ezmeral",
            icon="analytics",
            caption="install, configure and use",
        )
        .classes("w-full")
        .classes("text-bold") as ezmeral
    ):
        ui.label("Data Fabric").classes("text-bold")

        with ui.stepper().props("vertical").classes("w-full") as stepper:
            with ui.step("Configure", icon=""):
                df_prepare_ui()
                with ui.stepper_navigation():
                    ui.button(
                        "Setup",
                        icon="navigate_next",
                        on_click=lambda: dfprepare(stepper),
                    ).bind_enabled_from(
                        app.storage.user, "busy", backward=lambda x: not x
                    )
                    ui.button(
                        "Skip",
                        icon="skip_next",
                        on_click=stepper.next,
                    ).props("color=secondary")

            with ui.step("Install"):
                ui.input("Cluster Name", placeholder="ezlab").bind_value(
                    app.storage.general[DF], "cluster_name"
                ).classes("w-full")

                install_on_hosts = ui.input(
                    "Hosts",
                    placeholder="One or more comma-separated IPs, i.e., 10.1.1.x,10.1.1.y",
                    value=",".join([h["ip"] for h in app.storage.general[DF]["hosts"]]),
                ).classes("w-full")

                ui.input(
                    "Disks",
                    placeholder="One or more comma-separated raw disk path, i.e., /dev/sdb,/dev/vdb",
                ).bind_value(app.storage.general[DF], "maprdisks").classes("w-full")

                with ui.stepper_navigation():
                    ui.button(
                        "Install",
                        icon="navigate_next",
                        on_click=lambda: dfinstall(install_on_hosts.value, stepper),
                    ).bind_enabled_from(
                        app.storage.user, "busy", backward=lambda x: not x
                    )
                    ui.button(
                        "Skip",
                        icon="skip_next",
                        on_click=stepper.next,
                    ).props("color=secondary")
                    ui.button("Back", on_click=stepper.previous).props("flat")

            with ui.step("Cross-Cluster"):
                ui.label("Configure two clusters to join global namespace")
                ui.input(
                    "Local Cluster CLDB node", placeholder="core.ez.lab"
                ).bind_value(app.storage.general[DF], "crosslocalcldb").classes(
                    "w-full"
                )
                ui.input(
                    "Remote Cluster CLDB node", placeholder="edge.ez.lab"
                ).bind_value(app.storage.general[DF], "crossremotecldb").classes(
                    "w-full"
                )
                ui.input("Cluster Admin User", placeholder="mapr").bind_value(
                    app.storage.general[DF], "crossadminuser"
                ).classes("w-full")
                ui.input(
                    "Cluster Admin Password",
                    placeholder="mapr",
                    password=True,
                    password_toggle_button=True,
                ).bind_value(app.storage.general[DF], "crossadminpassword").classes(
                    "w-full"
                )

                with ui.stepper_navigation():
                    ui.button(
                        "Setup",
                        icon="navigate_next",
                        on_click=lambda: dfcrosscluster(stepper),
                    ).bind_enabled_from(
                        app.storage.user, "busy", backward=lambda x: not x
                    )
                    ui.button(
                        "Skip",
                        icon="skip_next",
                        on_click=stepper.next,
                    ).props("color=secondary")
                    ui.button("Back", on_click=stepper.previous).props("flat")

            with ui.step("Client"):
                ui.label("Client setup for DF")
                with ui.stepper_navigation():
                    ui.button(
                        "Connect",
                        icon="navigate_next",
                        on_click=lambda: dfclient(stepper),
                    ).bind_enabled_from(
                        app.storage.user, "busy", backward=lambda x: not x
                    )
                    ui.button("Back", on_click=stepper.previous).props("flat")

    ezmeral.bind_value(app.storage.general["ezui"], "ezmeral")


@ui.refreshable
def settings_menu():
    with (
        ui.expansion(
            "Settings",
            icon="settings",
            caption="for your environment",
        )
        .classes("w-full")
        .classes("text-bold") as settings
    ):
        ui.checkbox("Dry Run", value=False).bind_value(
            app.storage.general["config"], "dryrun"
        )
        ui.label("Network Settings").classes("text-bold")
        with ui.row():
            ui.input("VM Network", placeholder="10.1.1.0/24").bind_value(
                app.storage.general["config"], "cidr"
            )
            ui.input("Gateway", placeholder="10.1.1.1").bind_value(
                app.storage.general["config"], "gateway"
            )
            ui.input("Name Server", placeholder="10.1.1.1").bind_value(
                app.storage.general["config"], "nameserver"
            )
            ui.input("Domain", placeholder="ez.lab").bind_value(
                app.storage.general["config"], "domain"
            )
            ui.input("HTTP Proxy", placeholder="").bind_value(
                app.storage.general["config"], "proxy"
            )

        ui.label("Repositories").classes("text-bold")
        with ui.row().classes("w-full"):
            # YUM Repo
            ui.input("YUM Repo", placeholder="").bind_value(
                app.storage.general["config"], "yumrepo"
            )

            ui.input("EPEL Repo", placeholder="").bind_value(
                app.storage.general["config"], "epelrepo"
            )

        # MapR Repository Configuration
        with ui.row().classes("w-full"):
            ui.label("MapR Repo").classes("self-center")
            switch = (
                ui.switch("Use local", value=False)
                .props("left-label")
                .bind_value(app.storage.general["config"], "maprrepoislocal")
            )

            # Using default HPE repository
            with (
                ui.row()
                .bind_visibility_from(switch, "value", lambda x: not x)
                .classes("w-full items-justify")
            ):
                ui.input("HPE Passport e-mail").bind_value(
                    app.storage.general["config"], "maprrepouser"
                )
                ui.input(
                    "HPE Passport token", password=True, password_toggle_button=True
                ).bind_value(app.storage.general["config"], "maprrepotoken")

            # Using local repository
            with ui.row().bind_visibility_from(switch, "value"):
                ui.input(
                    "MapR Repo",
                    placeholder="https://repo.ez.lab/mapr/",
                ).bind_value(app.storage.general["config"], "maprlocalrepo")
                authlocal = ui.switch("Authenticate", value=True).bind_value(
                    app.storage.general["config"], "maprlocalrepoauth"
                )
                with ui.row():
                    ui.input("Username", password=True).bind_value(
                        app.storage.general["config"], "maprlocalrepousername"
                    ).bind_visibility(authlocal, "value")
                    ui.input("Password", password=True).bind_value(
                        app.storage.general["config"], "maprlocalrepopassword"
                    ).bind_visibility(authlocal, "value")

        ui.label("Cloudinit Settings").classes("text-bold")
        with ui.row():
            ui.input("username", placeholder="ezmeral").bind_value(
                app.storage.general["config"], "username"
            )
            ui.input("password", password=True, password_toggle_button=True).bind_value(
                app.storage.general["config"], "password"
            )

        ui.label("SSH Key").classes("text-bold")
        ssh_keys_ui()

    settings.bind_value(app.storage.general["ezui"], "settings")


# Business logic
async def connect(params: set):
    target, host, username, password = params
    try:
        if target == PVE:
            return await asyncio.to_thread(pve.connect, host, username, password)
        # elif target == KVM:
        #     return await asyncio.to_thread(kvm.connect, host, username, password)
        # elif target == VMWARE:
        #     return await asyncio.to_thread(vmw.connect, host, username, password)

    except Exception as error:
        ui.notify(error, type="negative")


async def new_vm():

    if app.storage.general["target"]["hve"] == PVE:
        try:
            app.storage.user["busy"] = True

            templates = [
                x
                for x in pve.vms(ez.connection)
                if x["template"] and x["type"] == "qemu"
            ]
            networks = pve.networks(ez.connection)
            storage = pve.storage(ez.connection)

            app.storage.user["busy"] = False

            if "template" not in app.storage.user.keys():
                app.storage.user["template"] = templates[0]

            def set_template(name):
                app.storage.user["template"] = [
                    t for t in templates if t["name"] == name
                ].pop()

            # select dialog
            with ui.dialog() as dialog, ui.card():
                t = (
                    ui.select(
                        options=[x["name"] for x in templates],
                        label="Template",
                        on_change=lambda e: set_template(e.value),
                    )
                    .classes("w-full")
                    .props("inline")
                    .bind_value(app.storage.general[PVE], "template")
                )
                s = (
                    ui.select(
                        options=[
                            x["storage"]
                            for x in storage
                            if x["node"] == app.storage.user["template"]["node"]
                        ],
                        label="Storage",
                    )
                    .classes("w-full")
                    .props("inline")
                    .bind_value(app.storage.general[PVE], "storage")
                )
                n = (
                    ui.select(
                        options=[
                            x["sdn"]
                            for x in networks
                            if x["node"] == app.storage.user["template"]["node"]
                        ],
                        label="Network",
                    )
                    .classes("w-full")
                    .props("inline")
                    .bind_value(app.storage.general[PVE], "network")
                )

                b = (
                    ui.select(
                        options=[
                            x["iface"]
                            for x in pve.bridges(
                                ez.connection, app.storage.user["template"]["name"]
                            )
                        ],
                        label="Bridge",
                    )
                    .classes("w-full")
                    .props("inline")
                    .bind_value(app.storage.general[PVE], "bridge")
                )

                e = (
                    ui.select(
                        options=[x["name"] for x in EZNODES],
                        label="Node Type",
                    )
                    .classes("w-full")
                    .props("inline")
                    .bind_value(app.storage.general[PVE], "eznode")
                )

                h = (
                    ui.input(label="VM Name", placeholder="ezvm")
                    .classes("w-full")
                    .props("inline")
                    .bind_value(app.storage.general[PVE], "hostname")
                )

                f = (
                    ui.input(
                        label="First IP",
                        placeholder=app.storage.general["config"]["gateway"],
                    )
                    .classes("w-full")
                    .props("inline")
                    .bind_value(app.storage.general[PVE], "firstip")
                )

                ui.button(
                    "Create",
                    on_click=lambda: dialog.submit(
                        (
                            next(
                                (
                                    # {"template": i}
                                    i
                                    for i in templates
                                    if i["name"] == t.value
                                ),
                                None,
                            ),
                            next(
                                (
                                    # {"storage": i}
                                    i
                                    for i in storage
                                    if i["storage"] == s.value
                                ),
                                None,
                            ),
                            # next(
                            #     (i for i in networks if i["sdn"] == n.value),
                            #     None,
                            # ),
                            b.value,  # Bridge
                            next(
                                (i for i in EZNODES if i["name"] == e.value),
                                None,
                            ),
                            h.value,  # hostname
                            f.value,  # firstip
                        )
                    ),
                )

            resources = await dialog
            if resources and all(resources):
                try:
                    vmcount = resources[3]["count"]

                    logger.info(f"Cloning {vmcount} VMs...")

                    queue = Queue(1000)
                    for count in range(vmcount):
                        # await run.io_bound(watch_clone_status, resources, n)
                        # run.io_bound(
                        pve.clone_vm(
                            proxmox=ez.connection,
                            resources=resources,
                            settings=dict(app.storage.general["config"]),
                            vm_number=0 if vmcount == 1 else count + 1,  # start from 1 not 0, use 0 for single vm
                            queue=queue,
                            dryrun=app.storage.general["config"].get("dryrun", True)
                        )
                    # reset counter to monitor completed tasks
                    app.storage.general["ezui"]["finished_tasks"] = 0

                    app.storage.user["busy"] = True
                    res = await run.io_bound(process_queue_messages, queue, vmcount)
                    if res:
                        ui.notify(f"Clone for {vmcount} VM(s) completed successfully", type="positive")

                    app.storage.user["busy"] = False
                    return True

                except Exception as error:
                    logger.error(error, exc_info=True)
                    app.storage.user["busy"] = False
                    return False

            else:
                ui.notify(f"Missing resources")

        except Exception as error:
            print("Query PVE error", error)
            ui.notify(f"PVE Error: {error}", type="negative")
            app.storage.user["busy"] = False

    else:
        ui.notify("not implemented")

    app.storage.user["busy"] = False
    return False


def process_queue_messages(queue: Queue, number_of_tasks: int = 1):
    while True:
        try:
            message = queue.get_nowait()
            if message is not None and TASK_FINISHED in message:
                app.storage.general["ezui"]["finished_tasks"] += 1
                logger.info(
                    f"Completed jobs: {app.storage.general['ezui']['finished_tasks']}"
                )
                if app.storage.general["ezui"]["finished_tasks"] == number_of_tasks:
                    logger.info("All tasks completed")
                    app.storage.general["ezui"]["finished_tasks"] = 0
                    queue.task_done()
                    return True
            else:
                if "error" in message or "failed" in message:
                    logger.error(message)
                    # ui.notify(message=message, type='warning', position='bottom-right')
                else:
                    logger.info(message)
            queue.task_done()
        except Empty:
            pass
        except Exception as error:
            queue.put(TASK_FINISHED)
            queue.task_done()
            print(error)


def connect_ui():
    # Login dialog
    with ui.dialog() as dialog, ui.card():
        t = (
            ui.radio(options=SUPPORTED_HVES)
            .classes("w-full")
            .props("inline")
            .bind_value(app.storage.general["target"], "hve")
        )

        h = (
            ui.input("Host", validation={"Required": nonzero})
            .classes("w-full")
            .bind_value(app.storage.general["target"], "host")
        )

        u = (
            ui.input("Username", validation={"Required": nonzero})
            .classes("w-full")
            .bind_value(app.storage.general["target"], "username")
        )

        p = (
            ui.input(
                "Password",
                password=True,
                password_toggle_button=True,
                validation={"Required": nonzero},
            )
            .classes("w-full")
            .bind_value(app.storage.general["target"], "password")
        )

        ui.button(
            "Login",
            on_click=lambda: dialog.submit((t.value, h.value, u.value, p.value)),
        )

    async def process_login():
        if app.storage.general["target"]["connected"]:
            # TODO: disconnect process/cleanup
            app.storage.general["target"]["connected"] = False
        else:
            result = await dialog
            if result:
                app.storage.user["busy"] = True
                connection = await connect(result)
                if connection:
                    app.storage.general["target"]["connected"] = True
                    ez.connection = connection
                else:
                    app.storage.general["target"]["connected"] = False
                    ui.notify(f"Get {connection} instead of a valid response")
                app.storage.user["busy"] = False

    with ui.row() as connection_status:
        ui.button(icon="cloud_done", on_click=process_login).bind_visibility_from(
            app.storage.general["target"],
            "connected",
        )
        ui.button(icon="cloud_off", on_click=process_login).bind_visibility_from(
            app.storage.general["target"],
            "connected",
            lambda x: not x,
        )
        # .bind_text_from(
        #     app.storage.general["target"],
        #     "connected",
        #     lambda x: "Disconnect" if x else "Connect",
        # )

    return connection_status


# Product specific UIs
async def dfprepare(pager: ui.stepper):
    hosts = app.storage.general[DF]["hosts"]
    # deep check list[dict] for missing values (if any host's name or IP is missing)
    if not (
        len(hosts) > 0
        and all(
            [values for values in [all(host) for host in [h.values() for h in hosts]]]
        )
    ):
        ui.notify("Missing values", type="warning")
        return False

    queue = Queue(1000)
    queue.put("Starting configuration for DF")
    for host in hosts:
        result = prepare_vm(
            hostname=host["name"],
            hostip=host["ip"],
            config=dict(app.storage.general["config"]),
            addhosts=[h for h in hosts if h["name"] != host["name"]],
            prepare_for=DF,
            queue=queue,
            dryrun=app.storage.general["config"]["dryrun"],
        )
        if app.storage.general["config"]["dryrun"]:
            get_doc(await result)
        elif result:
            pass
        else:
            ui.notify(f"{host} preparation failed!", type="warning")

    app.storage.user["busy"] = True
    res = await run.io_bound(process_queue_messages, queue, len(hosts))
    app.storage.user["busy"] = False

    if res:
        ui.notify(f"Finished with {len(hosts)} host(s)", type="positive")
        pager.next()
        return True
    else:
        return False


def df_prepare_ui():
    if not "hosts" in app.storage.general[DF].keys():
        app.storage.general[DF]["hosts"] = [{"name": "", "ip": ""}]
    with ui.row().classes("w-full items-center items-stretch") as prep_view:
        ui.label("Hosts")
        ui.space()
        with ui.row():

            def add_host():
                try:
                    app.storage.general[DF]["hosts"].append({"name": "", "ip": ""})
                except (KeyError, AttributeError):
                    app.storage.general[DF]["hosts"] = [{"name": "", "ip": ""}]
                finally:
                    add_remove_host.refresh(app.storage.general[DF]["hosts"])

            def remove_host():
                if len(app.storage.general[DF]["hosts"]) > 0:
                    app.storage.general[DF]["hosts"].pop()
                    add_remove_host.refresh()

            ui.button(icon="add_box", color=None, on_click=add_host).props("flat")
            ui.button(icon="delete", color=None, on_click=remove_host).props("flat")

    # Hosts
    add_remove_host(app.storage.general[DF]["hosts"])


async def dfinstall(hosts_string: str, pager: ui.stepper):
    queue = Queue(1000)
    queue.put("Starting DF cluster installation")

    result = cluster_install(
        hosts=hosts_string.split(","),
        config=dict(app.storage.general["config"]),
        queue=queue,
        name=app.storage.general[DF]["cluster_name"],
        disks=app.storage.general[DF]["maprdisks"],
        dryrun=app.storage.general["config"]["dryrun"],
    )

    if app.storage.general["config"]["dryrun"]:
        get_doc(await result)

    elif not result:
        ui.notify(f"Cluster creation failed!", type="negative")

    app.storage.user["busy"] = True
    res = await run.io_bound(process_queue_messages, queue, 1)
    app.storage.user["busy"] = False
    if res:
        ui.notify(f"Finished cluster creation", type="positive")
        pager.next()
        return True
    else:
        return False


async def dfcrosscluster(pager: ui.stepper):
    queue = Queue(1000)
    queue.put("Starting DF cross-cluster setup")

    if not crosscluster(
        config=dict(app.storage.general["config"]),
        settings=dict(app.storage.general[DF]),
        queue=queue,
    ):
        ui.notify(f"Cluster creation failed!", type="negative")

    app.storage.user["busy"] = True
    res = await run.io_bound(process_queue_messages, queue, 1)
    app.storage.user["busy"] = False
    if res:
        ui.notify(f"Finished cross-cluster setup", type="positive")
        pager.next()
        return True
    else:
        return False


async def dfclient(pager: ui.stepper):
    ui.notify("Not implemented", type='warning')
    return False

@ui.refreshable
def add_remove_host(hosts: list):
    if hosts is None:
        hosts = []
    for host in hosts:
        with ui.row().classes("w-full items-center no-wrap"):
            ui.input(
                "Hostname",
                placeholder="Host shortnames, ie, ezvm1,ezvm2...",
            ).bind_value(host, "name")
            ui.input(
                "IP Address",
                placeholder="10.1.1.x",
            ).bind_value(host, "ip")


def ua_ui(args):
    ui.notify(args)


#  async def vmware_ui(connection):
#     si = connection
#     vmtable = ui.table(
#         title="VMs",
#         columns=vmw.VM_COLUMNS,
#         rows=[],
#         selection="single",
#         on_select=lambda x: print(x),
#     ).classes("w-full max-h-200")

#     logger.info("waiting vms")

#     try:
#         for vm in await asyncio.to_thread(vmw.vms, si):
#             vmtable.add_rows([vm])

#     except Exception as e:
#         ui.notify(e, type="negative")

#     finally:
#         print("Done with vm")


def get_doc(content: dict):
    
    random_eof = "".join(random.choices(string.ascii_uppercase + string.digits, k=10))

    def indent(strings:str, spaces: int):
        return strings.strip().replace("\n", f"\n{' '*spaces}")

    settings = ""

    for key,value in content["settings"].items():
        settings += f"""
    #   {key}: {value if len(str(value)) < 80 else value[:77] + '...'}"""

    files = ""
    for filepath, filecontent in content["files"].items():
        files += f"""
    sudo cat <<{random_eof}>{filepath}
    {indent(filecontent,4)}
    {random_eof}
    """

    commands = ""
    for command in content["commands"]:
        commands += f"""
    {indent(command,4)}
    """

    doc = f"""
    #!/usr/bin/env bash

    # Ezlab Run {time.strftime('%Y%m%d%H%M%z')}
    
    # {content["task"]}

    # Settings
        {settings}

    # Write Files
        {files}
    
    # Run Commands
        {commands}

    # End of {content['task']}

"""

    with ui.dialog().props("full-width") as dialog, ui.card().classes("w-full"):
        ui.code(doc, language="bash").classes(
            "w-full text-wrap"
        )
        with ui.row():
            ui.button(
                "Save (useless on app mode)",
                icon="save",
                on_click=lambda: ui.download(doc.encode(), "ezlab-runbook.sh"),
            )
            ui.button(
                "Close",
                icon="cancel",
                on_click=dialog.close,
            )
    dialog.open()



from typing import Annotated
import typer
from .connect import connect_to_target

from .common import fail, target_classes, target_names

from . import pve
from . import vmw

app = typer.Typer(add_completion=True, no_args_is_help=True)


@app.command()
def vm(
    target: Annotated[target_names, typer.Option("--target", "-t", prompt=True)],
    host: Annotated[
        str, typer.Option("--host", "-h", envvar="EZLAB_HOST", prompt=True)
    ],
    username: Annotated[
        str, typer.Option("--username", "-u", envvar="EZLAB_USER", prompt=True)
    ],
    password: Annotated[
        str,
        typer.Option(
            "--password", "-p", prompt=True, hide_input=True, envvar="EZLAB_PASSWORD"
        ),
    ],
):
    connection = connect_to_target(target, host, username, password)

    vms = []
    connection_type = connection.__class__.__name__
    if connection_type == target_classes.pve.value:
        vms = pve.delete_vms(connection)  # type: ignore
    elif connection_type == target_classes.vmw.value:
        vms = vmw.delete_vms(connection)  # type: ignore
    else:
        fail("Unknown connection type")

    if vms:
        for vm in vms:
            print(f"{vm} is deleted.")


@app.command()
def template(
    target: Annotated[target_names, typer.Option("--target", "-t", prompt=True)],
    host: Annotated[
        str, typer.Option("--host", "-h", envvar="EZLAB_HOST", prompt=True)
    ],
    username: Annotated[
        str, typer.Option("--username", "-u", envvar="EZLAB_USER", prompt=True)
    ],
    password: Annotated[
        str,
        typer.Option(
            "--password", "-p", prompt=True, hide_input=True, envvar="EZLAB_PASSWORD"
        ),
    ],
):
    connection = connect_to_target(target, host, username, password)

    vms = []
    connection_type = connection.__class__.__name__
    if connection_type == target_classes.pve.value:
        vms = pve.delete_templates(connection)  # type: ignore
    elif connection_type == target_classes.vmw.value:
        vms = vmw.delete_templates(connection)  # type: ignore
    else:
        fail("Unknown connection type")

    for vm in vms:
        print(f"{vm} is deleted.")



from enum import Enum
import os

ini_file = os.path.expanduser("~/.ezconfig.ini")

class ez_product(str, Enum):
    df = "ezdf"
    ua = "ezua"
    generic = "generic"


ez_nodes = [
    {
        "name": "ua-control",
        "product": ez_product.ua.value,
        "cores": 4,
        "memGB": 8,
        "os_disk_size": 300,
        "data_disk_size": 500,
        "no_of_disks": 1,
        "count": 2,
    },
    {
        "name": "ua-worker",
        "product": ez_product.ua.value,
        "cores": 32,
        "memGB": 128,
        "os_disk_size": 300,
        "data_disk_size": 500,
        "no_of_disks": 2,
        "count": 3,
    },
    {
        "name": "df-singlenode",
        "product": ez_product.df.value,
        "cores": 32,
        "memGB": 96,
        "os_disk_size": 200,
        "data_disk_size": 500,
        "no_of_disks": 1,
        "count": 1,
    },
    {
        "name": "df-5nodes",
        "product": ez_product.df.value,
        "cores": 16,
        "memGB": 34,
        "os_disk_size": 200,
        "data_disk_size": 500,
        "no_of_disks": 1,
        "count": 5,
    },
    {
        "name": "client",
        "product": ez_product.generic.value,
        "cores": 2,
        "memGB": 4,
        "os_disk_size": 50,
        "no_of_disks": 0,
        "count": 1,
    },
    {
        "name": "generic",
        "product": ez_product.generic.value,
        "cores": 1,
        "memGB": 2,
        "os_disk_size": 20,
        "no_of_disks": 0,
        "count": 1,
    },
]


ez_files = {
    ez_product.df.value: {
        "/etc/sysctl.d/mapr.conf": "vm.swappiness=1\n",
        "/etc/profile.d/mapr.sh": "umask 0022\n",
    },
    ez_product.ua.value: {},
    ez_product.generic.value: {},
}

ez_commands = {
    ez_product.df.value: [
        # "sudo chown root:root /etc/sudoers.d/mapr -- 2>/dev/null && sudo chmod 440 /etc/sudoers.d/mapr",
        # disable subscription manager - TODO: needs testing
        "[ -f /etc/yum/pluginconf.d/product-id.conf ] && sudo sed -i 's/^enabled=0/enabled=1/' /etc/yum/pluginconf.d/product-id.conf",
        "[ -f /etc/yum/pluginconf.d/subscription-manager.conf ] && sudo sed -i 's/^enabled=0/enabled=1/' /etc/yum/pluginconf.d/subscription-manager.conf",
        "sudo sysctl vm.swappiness=1 > /dev/null",
        # "sudo systemctl disable --now numad",
        "sudo sed -i 's/^[^#]*PasswordAuthentication[[:space:]]no/PasswordAuthentication yes/' /etc/ssh/sshd_config",
        "sudo systemctl restart sshd",
        "sudo dnf install -y -q wget > /dev/null",
    ],
    ez_product.ua.value: [
        # disable subscription manager - TODO: needs testing
        "[ -f /etc/yum/pluginconf.d/product-id.conf ] && sudo sed -i 's/^enabled=0/enabled=1/' /etc/yum/pluginconf.d/product-id.conf",
        "[ -f /etc/yum/pluginconf.d/subscription-manager.conf ] && sudo sed -i 's/^enabled=0/enabled=1/' /etc/yum/pluginconf.d/subscription-manager.conf",
        "sudo dnf -q -y install nfs-utils policycoreutils-python-utils conntrack-tools jq tar >/dev/null",
        "sudo dnf --setopt=tsflags=noscripts install -y -q iscsi-initiator-utils >/dev/null",
        'echo "InitiatorName=$(/sbin/iscsi-iname)" | sudo tee -a /etc/iscsi/initiatorname.iscsi >/dev/null',
        "sudo systemctl enable --now iscsid 2>&1 > /dev/null",
        "sudo ethtool -K eth0 tx-checksum-ip-generic off >/dev/null",
        "sudo sed -i 's/^[^#]*PermitRootLogin[[:space:]]no/PermitRootLogin yes/' /etc/ssh/sshd_config",
        "sudo systemctl restart sshd",
    ],
    ez_product.generic.value: [
        "[ -f /etc/yum/pluginconf.d/product-id.conf ] && sudo sed -i 's/^enabled=0/enabled=1/' /etc/yum/pluginconf.d/product-id.conf",
        "[ -f /etc/yum/pluginconf.d/subscription-manager.conf ] && sudo sed -i 's/^enabled=0/enabled=1/' /etc/yum/pluginconf.d/subscription-manager.conf",
        "sudo sed -i 's/^[^#]*PasswordAuthentication[[:space:]]no/PasswordAuthentication yes/' /etc/ssh/sshd_config",
        "sudo systemctl restart sshd",
    ],
}

ez_installers = {
    ez_product.df.value: {
        "image": "maprtech/edf-seed-container:latest",
        "start_installer": "./datafabric_container_setup.sh",
        "url": "https://localhost:8443/app/dfui/#/installer",
        "commands": [],
    },
    ez_product.ua.value: {
        "start_installer": "./start_ezua_installer_ui.sh",
        # "image": "marketplace.us1.greenlake-hpe.com/ezua/ezua/hpe-ezua-installer-ui:latest",
        "image": "us.gcr.io/mapr-252711/hpe-ezua-installer-ui:1.2.0-d16afb0",
        "url": "http://localhost:8080",
    },
    ez_product.generic.value: {},
}



from .common import fail, target_classes, target_names
from . import pve
from . import vmw


def connect_to_target(target, host, username, password):
    """Create connection to target API

    Args:
        target (enum): pve or vmw
        host (string): API host/IP to connect
        username (str): API user, required permissions ...
        password (str): API user password

    Returns:
        class: API connection class
    """
    connection = None
    if target == target_names.pve:
        connection = pve.get_connection(host, username, password)

    elif target == target_names.vmw:
        connection = vmw.get_connection(host, username, password)

    else:
        fail("Unknown target")

    return connection


def select_vms(connection):
    """Returns all VMs from connection

    Args:
        connection (class<API>): connection class returned by the API connection

    Returns:
        list<class>: List of all VMs returned by API call
    """
    connection_type = connection.__class__.__name__
    if connection_type == target_classes.pve.value:
        return pve.select_vms(connection)
    elif connection_type == target_classes.vmw.value:
        return vmw.get_all_vms(connection)
    else:
        fail("Unknown connection type")
    return None


# def get_vm_ip(connection, vm):
#     connection_type = connection.__class__.__name__
#     if connection_type == target_classes.pve.value:
#         return pve.get_vm_ip(connection, vm)
#     elif connection_type == target_classes.vmw.value:
#         fail("Vmware not implemented yet")
#     else:
#         fail("Unknown connection type")
#     return None



import asyncio
import base64
import io
from ipaddress import IPv4Address, ip_address
import os
from queue import Queue
import random
import re
import shlex
import socket
import string
import subprocess
import cryptography
from cryptography.hazmat.primitives import serialization


# from threading import Thread
from time import sleep
from paramiko import (
    AuthenticationException,
    AutoAddPolicy,
    BadHostKeyException,
    RSAKey,
    SSHException,
    SSHClient,
)

from ezlab.parameters import *


def nonzero(val):
    return len(val) != 0


def fire_and_forget(f):
    def wrapped(*args, **kwargs):
        return asyncio.get_event_loop().run_in_executor(
            None, f, *args, *[v for v in kwargs.values()]
        )

    return wrapped


def wait_for_ssh(host, username, privatekey):
    client = SSHClient()
    client.set_missing_host_key_policy(AutoAddPolicy)
    try:
        while True:
            client.connect(
                host,
                username=username,
                pkey=RSAKey.from_private_key(io.StringIO(privatekey)),
            )
            break
    except (
        BadHostKeyException,
        AuthenticationException,
        ConnectionResetError,
        SSHException,
        socket.error,
    ):
        print(f"{host} still waiting ssh...")
        sleep(5)
    finally:
        print(f"{host} connected")
        client.close()

    return True


def ssh_run_command(
    host: str,
    username: str,
    keyfile: str,
    command: str,
):
    """
    Run commands over ssh using default credentials
    Params:
        <host>              : hostname/IP to SSH
        <username>          : username to authenticate
        <keyfile>           : private key for <username>
        <command>           : command to run
    Returns:
        Generator with stdout. Calling function should process output with a for loop.
        Prints out stderr directly into terminal.

    """
    client = SSHClient()
    client.set_missing_host_key_policy(AutoAddPolicy)
    command_output = None
    # print(f"Try connection to {host}")
    try:
        client.connect(
            host,
            username=username,
            pkey=RSAKey.from_private_key(io.StringIO(keyfile)),
            timeout=60,
        )

        if "hostname -f" not in command:
            yield f"{host} running: {command}"

        _stdin, _stdout, _stderr = client.exec_command(command, get_pty=True)
        command_errors = _stderr.read().decode().strip()
        command_output = _stdout.read().decode().strip()

        if command_errors and command_errors != "":
            yield f"[CMDERR]: {command_errors}"
        if command_output and command_output != "":
            # should return only the command output
            yield command_output

    except Exception as error:
        print("SSH run exception: ", error)

    finally:
        client.close()

    # return command_output


def ssh_content_into_file(
    host: str, username: str, keyfile: str, content: str, filepath: str
):
    random_eof = "".join(random.choices(string.ascii_uppercase + string.digits, k=10))
    command = f"""cat << {random_eof} | sudo tee {filepath}
{content}
{random_eof}
"""
    return ssh_run_command(
        host=host,
        username=username,
        keyfile=keyfile,
        command=command,
    )


def find_app(name: str):
    """Return executable path for `name` on PATH or in CWD."""
    from shutil import which

    return which(cmd=name, path=f"{os.environ.get('PATH')}:{os.getcwd()}")


def execute(cmd: str):
    proc = subprocess.Popen(
        args=shlex.split(cmd),
        stdout=subprocess.PIPE,
        universal_newlines=True,
        # env=sub_env,
    )

    while proc.stdout:
        output = proc.stdout.readline().strip()
        if output == "" and proc.poll() is not None:
            break
        if output:
            yield f">>> {output}"
    rc = proc.poll()


def toB64(string: str):
    return base64.b64encode(string.encode("ascii")).decode("utf-8")


def validate_ip(ip_string: str) -> bool:
    try:
        return isinstance(ip_address(ip_string), IPv4Address)
    except:
        return False


def get_fqdn(ip_string: str) -> str:
    fqdn, _, _ = socket.gethostbyaddr(ip_string)
    return fqdn


@fire_and_forget
def prepare_vm(
    hostname: str,
    hostip: str,
    config: dict,
    addhosts: list,
    prepare_for: str,
    queue: Queue,
    dryrun: bool,
):
    """
    Configure VMs for selected product, taking care of pre-requisites
    Params:
        <hostname>      : VM hostname
        <hostip>        : VM IP address (should be reachable)
        <config>        : Dictinary of configuration from app settings
        <addhosts>      : Other hosts to add to /etc/hosts file
        <prepare_for>   : EZUA | EZDF | Generic (from parameters: UA|DF|GENERIC)
        <queue>         : Message queue
        <dryrun>        : Do not execute but report

    Returns <bool>     : True in succesful completion, False otherwise. If dryrun, returns dict of job run
    """

    response = {}
    if dryrun:
        response["task"] = (
            f"Prepare {hostname} ({hostip}) for {prepare_for} {'with hosts: ' + ','.join(addhosts) if len(addhosts) > 0 else 'as single node'}"
        )
        response["settings"] = config

    domain = config["domain"]
    cidr = config["cidr"]
    username = config["username"]
    # password = config["password"]
    keyfile = config["privatekeyfile"]
    # commands to execute, always prepend with generic/common commands, such as enable password auth, disable subscription mgr etc.
    commands = list(SETUP_COMMANDS[GENERIC]) + list(SETUP_COMMANDS[prepare_for])
    after_generic_commands = len(list(SETUP_COMMANDS[GENERIC]))

    # to add to /etc/hosts
    otherhosts = "\n".join(
        [f"{h['ip']} {h['name']}.{domain} {h['name']}" for h in addhosts]
    )

    files = {
        # "/etc/cloud/templates/hosts.redhat.tmpl": f"127.0.0.1 localhost.localdomain localhost\n{hostip} {hostname}.{domain} {hostname}\n"
        "/etc/hosts": f"127.0.0.1 localhost.localdomain localhost\n{hostip} {hostname}.{domain} {hostname}\n{otherhosts}"
    }

    noproxy = (
        f"{hostip},{re.split(':|/', config.get('maprlocalrepo').split('://')[1])[0]}"
        if config.get("maprrepoislocal", False)
        else hostip
    )

    # update env files for proxy
    if config.get("proxy", "").strip() != "":
        environment_file = ""
        proxy_file = ""
        for key, value in get_proxy_vars(config["proxy"]).items():
            val = value.format(vm_domain=domain, vm_network=cidr, no_proxy=noproxy)
            environment_file += f"{key}={val}\n"
            proxy_file += f"export {key}={val}\n"

        files.update({"/etc/environment": environment_file})
        files.update({"/etc/profile.d/proxy.sh": proxy_file})

    # TODO: Not tested
    if (
        config.get("airgap_registry", "").strip() != ""
        and config["airgap_registry"].split("://")[0] == "http"
    ):
        files.update(
            {
                "/etc/docker/daemon.json": get_insecure_registry(
                    config["airgap_registry"]
                )
            }
        )

    repo_content = ""
    if config.get("yumrepo", "").strip() != "":
        repo_content = get_repo_content(config["yumrepo"])
    else:
        print("Using system repositories")

    if config.get("epelrepo", "").strip() != "":
        repo_content += f"""
[local-epel]
name = Local EPEL
enabled = 1
gpgcheck = 0
baseurl = {config.get("epelrepo")}
priority=1
proxy=

"""
    else:
        epel_default = "subscription-manager repos --enable codeready-builder-for-rhel-8-x86_64-rpms; sudo dnf install -y https://dl.fedoraproject.org/pub/epel/epel-release-latest-8.noarch.rpm"
        commands.insert(after_generic_commands, epel_default)

    if repo_content != "": # yum or epel repo given
        files.update({"/etc/yum.repos.d/local.repo": repo_content})

    if dryrun:
        response["files"] = files
    else:
        try:
            for filepath, content in files.items():
                print(f"{hostip} Writing {filepath}")
                for result in ssh_content_into_file(
                    host=hostip,
                    username=username,
                    keyfile=keyfile,
                    content=content,
                    filepath=filepath,
                ):
                    queue.put(f"{hostip} COPY {filepath}, CONTENT:\n{result}")

        except Exception as error:
            print(error)
            return False

    if (
        config.get("yumrepo", "").strip() != ""
        or config.get("epelrepo", "").strip() != ""
    ):
        commands.insert(
            after_generic_commands, "sudo dnf config-manager --set-disabled baseos appstream extras"
        )

    if config.get("proxy", "").strip() != "":
        commands.insert(
            after_generic_commands + 1,
            # f"echo proxy={config['proxy']} | sudo tee -a /etc/dnf/dnf.conf > /dev/null",
            "sudo sed -i '/proxy=/d' /etc/dnf/dnf.conf > /dev/null; echo proxy={proxy_line} | sudo tee -a /etc/dnf/dnf.conf".format(
                proxy_line=(
                    config["proxy"]
                    if config["proxy"][-1] == "/"
                    else config["proxy"] + "/"
                )
            ),
        )

    # rename host
    commands.insert(after_generic_commands + 2, f"sudo hostnamectl set-hostname {hostname}.{domain}")

    if dryrun:
        response["commands"] = commands

    else:
        try:
            for command in commands:
                print(f"{hostip} Running: {command}")
                for output in ssh_run_command(
                    host=hostip, username=username, keyfile=keyfile, command=command
                ):
                    queue.put(f"{hostip} SSH: {output}")
        except Exception as error:
            print(error)
            return False

    queue.put(TASK_FINISHED)

    return response if dryrun else True


def get_proxy_vars(proxy):
    if proxy[-1] != "/":
        proxy += "/"
    return {
        "HTTP_PROXY": proxy,
        "http_proxy": proxy,
        "HTTPS_PROXY": proxy,
        "https_proxy": proxy,
        "NO_PROXY": NO_PROXY,
        "no_proxy": NO_PROXY,
    }


def get_insecure_registry(registry):
    return f"""
{
  "log-driver": "journald",
  "insecure-registries" : [ {registry} ]
}
"""


def get_repo_content(repo_url: str):
    repo_content = ""
    for repo in ["BaseOS", "AppStream", "PowerTools", "extras"]:
        repo_content += fr"""
[local-{repo.lower()}]
name = Local {repo}
enabled = 1
gpgcheck = 0
baseurl = {repo_url.rstrip('/')}/\$releasever/{repo}/\$basearch/os
ui_repoid_vars = releasever basearch
priority=1
proxy=
"""
    return repo_content


def is_file(file):
    return os.path.isfile(os.path.abspath(file))


def get_privatekey_from_string(key):
    try:
        # try OpenSSL format (RSA?)
        return serialization.load_pem_private_key(key.encode(), password=None)
    except ValueError:
        # try OpenSSH format
        try:
            return serialization.load_ssh_private_key(key.encode(), password=None)
        except (ValueError, TypeError, cryptography.exceptions.UnsupportedAlgorithm):
            return None
    except Exception as error:
        print(error)
        return None


def get_opensshpub(privatekey):
    try:
        return (
            privatekey.public_key()
            .public_bytes(
                encoding=serialization.Encoding.OpenSSH,
                format=serialization.PublicFormat.OpenSSH,
            )
            .decode()
        )
    except Exception as error:
        print(error)
        return None


def debug_queue_to_here(queue, message):
    print(message)
    queue.put(TASK_FINISHED)
    return True



import logging
# from pyVmomi import vim
# from pyVim.connect import SmartConnect, Disconnect
# from .vmw_pchelper import *

logger = logging.getLogger(__name__)

# Constants
VM_COLUMNS = [
    {
        "name": "name",
        "label": "Name",
        "field": "name",
        "align": "left",
        "sortable": True,
    },
    # {"name": "datacenter", "label": "Datacenter", "field": "datacenter", "align": "left"},
    # {"name": "host", "label": "Host", "field": "node", "align": "left"},
    # {"name": "folder", "label": "Folder", "field": "folder", "align": "left"},
    {"name": "status", "label": "Status", "field": "status", "align": "left"},
    {"name": "template", "label": "Template", "field": "template", "align": "left"},
]

DC_COLUMNS = [
    {
        "name": "id",
        "label": "ID",
        "field": "id",
        "align": "left",
        "sortable": True,
    },
    {
        "name": "name",
        "label": "Name",
        "field": "name",
        "align": "left",
        "sortable": True,
    },
]


# Shamelessly borrowed from:
# https://github.com/vmware/pyvmomi-community-samples/tree/master/samples
# def connect(host: str, username: str, password: str):
#     try:
#         return SmartConnect(
#             host=host, user=username, pwd=password, disableSslCertValidation=True
#         )
#         # atexit.register(Disconnect, si)
#     except Exception as e:
#         return e


# def disconnect(si):
#     Disconnect(si)


# def datacenters(si):
#     # return si.RetrieveContent().rootFolder.childEntity
#     content = si.RetrieveContent()
#     dclist = get_all_obj(content, [vim.Datacenter])
#     return [{"id": str(key).strip("'"), "name": dclist[key]} for key in dclist.keys()]


# def vms(si):
#     # for folder in get_all_obj(content, [vim.Folder]):
#     vms = get_all_obj(si.content, [vim.VirtualMachine])
#     for vm in vms:
#         # print(vm.config.name)
#         yield {
#             "id": str(vm).strip("'"),
#             "name": vm.config.name,
#             "config": vm.config,
#             "template": vm.config.template,
#             "parent": vm.parent,
#         }

#     # return [ { 'id': str(key).strip("'"), 'name': vms[key], "config": vms[key].config, "template": vms[key].config.template, "parent": vms[key].parent } for key in vms.keys()]



# ### OLD CODE BELOW

# def select_from_objects(entities, title="Select"):
#     return select_from_list(
#         title=title,
#         list=sorted(entities, key=lambda x: x.name),
#         name_generator=lambda x: x.name,
#     )


# def wait_task_state(task: vim.Task):
#     try:
#         WaitForTask(task)
#     except Exception as e:
#         fail(e)
#     return task.info.state


# def wait_task_result(task: vim.Task):
#     WaitForTask(task)
#     return task.info.result


# def get_resources(si: vim.ServiceInstance):
#     resources = {}
#     content: vim.ServiceInstanceContent = si.RetrieveContent()

#     datacenter: vim.Datacenter = content.rootFolder.childEntity[0]

#     # hostfolder: vim.Folder = datacenter.hostFolder.childEntity[0]

#     # host: vim.HostSystem = hostfolder.host[0]

#     # network: vim.Network = host.network[0]

#     # datastore: vim.Datastore = host.datastore[0]

#     # vm_folder: vim.Folder = datacenter.vmFolder.childEntity[0]

#     # # Handle esxi - no vm_folder childEntity
#     # if isinstance(vm_folder, vim.VirtualMachine):
#     #     vm_folder = datacenter.vmFolder

#     # virtual_machines = [v for v in vms(vm_folder=vm_folder)]
#     # # can use any vm as template, since esxi doesn't support template
#     # # [v for v in get_all_vms(vm_folder=vm_folder) if v.config.template], "Template"

#     # resources.update(
#     # {
#     # "datacenter": datacenter,
#     # "host": host,
#     # "network": network,
#     # "datastore": datastore,
#     # "vm_folder": vm_folder,
#     # "vms": virtual_machines,
#     # }
#     # )

#     return resources


# def create_template(si: vim.ServiceInstance):
#     content: vim.ServiceInstanceContent = si.RetrieveContent()

#     datacenter: vim.Datacenter = select_from_objects(
#         content.rootFolder.childEntity, "Data Center"
#     )

#     hostfolder: vim.Folder = select_from_objects(
#         datacenter.hostFolder.childEntity, "Folder"
#     )

#     host: vim.HostSystem = select_from_objects(hostfolder.host, "Template Host")

#     network: vim.Network = select_from_objects(host.network, "Network")

#     datastore: vim.Datastore = select_from_objects(host.datastore, "Tempate Datastore")

#     vmdk = select_from_list(
#         title="Base OS Image (vmdk)",
#         list=[file_in_ds for file_in_ds in find_vmdks_in_datastore(datastore)],
#         name_generator=lambda x: "".join([x["folderPath"], x["file"].path]),
#     )

#     if vmdk is None:
#         fail("No template disk!")

#     vm_folder: vim.Folder = select_from_objects(
#         datacenter.vmFolder.childEntity, "VM Folder"
#     )
#     # Handle esxi - no vm_folder childEntity
#     if isinstance(vm_folder, vim.VirtualMachine):
#         vm_folder = datacenter.vmFolder

#     # Create template with "generic" host definition
#     ez_node = [node for node in ez_nodes if node["name"] == "generic"][0]

#     if "rhel" in vmdk["file"].path.lower():
#         vm_name = "rhel-template1"
#     else:
#         vm_name = "rocky-template1"

#     # resolve name conflict
#     all_vms = get_all_vms(vm_folder)

#     if vm_name in [vm.name for vm in all_vms]:
#         vm_name = re.sub(
#             "(\d+)(?!.*\d)",
#             lambda x: str(int(x.group(0)) + 1),
#             all_vms[-1].name,
#         )

#     print(f"Creating {vm_name}")
#     try:
#         vm = create_vm(
#             vm_folder=vm_folder,
#             vm_name=vm_name,
#             datastore=datastore,
#             host=host,
#             ez_node=ez_node,
#         )

#         if vm is not None:
#             if not add_scsi_controller(vm):
#                 fail(f"Failed at adding controller to {vm.name}")

#             # source_disk_path = f"{vmdk['folderPath']} {vmdk['file'].path}"
#             # os_disk_path = f"{vmdk['folderPath']} {vm_name}/{vm_name}.vmdk"

#             print(f"Adding disk... ", end="")

#             if not add_os_disk_to_vm(
#                 vm=vm,
#                 vmdk=vmdk,
#                 disk_sizeGB=ez_node["os_disk_size"],
#             ):
#                 fail(f"Failed a adding os disk to {vm.name}")

#             print(f"Adding vNIC... ", end="")

#             if not add_nic_on_network(vm=vm, network=network):
#                 fail(f"Failed at adding vNIC to VM {vm.name}")
#             print(network.name)

#             try:
#                 print("Convert to template.")
#                 vm.MarkAsTemplate()
#             except vmodl.fault.NotSupported:
#                 print("ignoring...")
#                 pass
#             yield vm_name
#         else:
#             fail(f"Failed to create {vm_name}")

#     except Exception as error:
#         print(f"failed at configuring {vm.name}")
#         fail(error)


# def find_vmdks_in_datastore(datastore):
#     spec = vim.host.DatastoreBrowser.SearchSpec(
#         sortFoldersFirst=True, query=[vim.host.DatastoreBrowser.VmDiskQuery()]
#     )
#     spec.searchCaseInsensitive = False
#     spec.matchPattern = ["*.vmdk"]
#     task = datastore.browser.SearchSubFolders(f"[{datastore.name}]", spec)
#     for res in wait_task_result(task):
#         for file in res.file:
#             yield {
#                 "datastore": res.datastore,
#                 "file": file,
#                 "folderPath": res.folderPath,
#             }


# def copy_file(content, datacenter, source, destination):
#     print(f"Copying {source} to {destination}... ", end="")

#     print(
#         wait_task_state(
#             content.fileManager.CopyFile(
#                 sourceName=source,
#                 sourceDatacenter=datacenter,
#                 destinationName=destination,
#                 destinationDatacenter=datacenter,
#             )
#         )
#     )


# def clone_vms(connection: vim.ServiceInstance, customisations: dict, nodes: list):
#     resources = get_resources(connection)

#     name_index = 0  # index for vm names
#     threads = list()

#     for ez_node in nodes:
#         for idx in range(ez_node["count"]):
#             name_index += 1
#             vm_ip = customisations["first_vm_ip"]
#             if vm_ip != "dhcp":
#                 # vm number starting at 1, ip index should start with 0
#                 vm_ip = str(
#                     ip_address(customisations["first_vm_ip"]) + (name_index - 1)
#                 )

#             # yield clone_vm(
#             #     vm_name=customisations["vmname"] + str(name_index),
#             #     vm_ip=vm_ip,
#             #     ez_node=ez_node,
#             #     resources=resources,
#             # )

#             task = ReturningThread(
#                 target=clone_vm,
#                 kwargs={
#                     "vm_name": customisations["vmname"] + str(name_index),
#                     "vm_ip": vm_ip,
#                     "ez_node": ez_node,
#                     "resources": resources,
#                 },
#             )
#             threads.append(task)

#     start_tasks_delayed(threads, 2)

#     return wait_tasks(threads)


# def clone_vm(vm_name: str, vm_ip: str, ez_node: dict, resources: dict):
#     """
#     Clone a VM from a template/VM

#     """

#     # datacenter: vim.Datacenter = resources["datacenter"]
#     datastore: vim.Datastore = resources["datastore"]
#     # network: vim.Network = resources["network"]
#     host: vim.HostSystem = resources["host"]
#     vm_folder: vim.Folder = resources["vm_folder"]
#     template_vm: vim.VirtualMachine = resources["template_vm"]

#     defaults = config.get()
#     # vm_netmask = defaults["NETWORK"]["vm_cidr"].split("/")[1]
#     vm_netmask = str(ip_network(defaults["NETWORK"]["vm_cidr"]).netmask)
#     vm_gateway = defaults["NETWORK"]["vm_gateway"]
#     vm_domain = defaults["NETWORK"]["vm_domain"]
#     vm_nameserver = defaults["NETWORK"]["vm_nameserver"]

#     # set relospec
#     relospec = vim.vm.RelocateSpec()
#     relospec.datastore = template_vm.datastore[0]
#     relospec.pool = host.parent.resourcePool

#     nic = vim.vm.device.VirtualDeviceSpec()
#     nic.operation = vim.vm.device.VirtualDeviceSpec.Operation.edit
#     nic.device = get_net_device(vm=template_vm)
#     nic.device.addressType = "assigned"
#     nic.device.connectable = vim.vm.device.VirtualDevice.ConnectInfo()
#     nic.device.connectable.startConnected = True
#     nic.device.connectable.allowGuestControl = True

#     guest_map = vim.vm.customization.AdapterMapping()
#     guest_map.adapter = vim.vm.customization.IPSettings()
#     if vm_ip == "dhcp":
#         guest_map.adapter.ip = vim.vm.customization.DhcpIpGenerator()
#     else:
#         guest_map.adapter.ip = vim.vm.customization.FixedIp()
#         guest_map.adapter.ip.ipAddress = vm_ip
#         guest_map.adapter.subnetMask = vm_netmask
#         guest_map.adapter.gateway = vm_gateway
#         guest_map.adapter.dnsDomain = vm_domain
#         guest_map.adapter.dnsServerList = [vm_nameserver]

#     # VM config spec
#     vmconf = new_vm_spec(
#         datastore=template_vm.datastore[0], ez_node=ez_node, name=vm_name
#     )
#     vmconf.deviceChange = [nic]

#     # DNS settings
#     globalip = vim.vm.customization.GlobalIPSettings()
#     globalip.dnsServerList = [vm_nameserver]
#     globalip.dnsSuffixList = [vm_domain]

#     # Hostname settings
#     # ident = vim.vm.customization.LinuxPrep()
#     # ident.domain = vm_domain
#     # ident.hostName = vim.vm.customization.FixedName()
#     # ident.hostName.name = vm_name
#     # ident.timeZone = "Europe/London"

#     cloudinit_spec = vim.vm.customization.CloudinitPrep()
#     ssh_pubkey = ""
#     with open(os.path.expanduser(defaults["CLOUDINIT"]["ssh_keyfile"])) as f:
#         ssh_pubkey = f.read()

#     password_hash = crypt.crypt(defaults["CLOUDINIT"]["vm_password"])

#     cloudinit_spec.userdata = get_userdata(
#         username=defaults["CLOUDINIT"]["vm_username"],
#         password=password_hash,
#         ssh_pubkey=ssh_pubkey,
#         vm_name=vm_name,
#         vm_domain=vm_domain,
#     )
#     # TODO: Handle dhcp assigned IP
#     cloudinit_spec.metadata = get_metadata(
#         vm_name=vm_name,
#         vm_ip=vm_ip,
#         vm_netmask=vm_netmask,
#         vm_gateway=vm_gateway,
#         vm_nameserver=vm_nameserver,
#         vm_domain=vm_domain,
#     )

#     print(cloudinit_spec)

#     customspec = vim.vm.customization.Specification()
#     customspec.nicSettingMap = [guest_map]
#     customspec.globalIPSettings = globalip
#     customspec.identity = cloudinit_spec

#     clonespec = vim.vm.CloneSpec()
#     clonespec.location = relospec
#     clonespec.config = vmconf
#     clonespec.customization = customspec
#     clonespec.powerOn = False
#     clonespec.template = False

#     print(clonespec)

#     print(f"{vm_name} cloning... ", end="")
#     try:
#         task = template_vm.Clone(folder=vm_folder, name=vm_name, spec=clonespec)
#         vm: vim.VirtualMachine = wait_task_result(task)
#         print("ok")
#     except vmodl.fault.NotSupported:
#         fail("esxi doesn't support vm clones, please use vcenter")

#     # add new disks /// assume no_of_disks are 0 or 1 or 2
#     if ez_node["no_of_disks"] > 0:
#         for i in range(ez_node["no_of_disks"]):
#             add_new_disk_to_vm(
#                 vm=vm, datastore=datastore, disk_size=ez_node["data_disk_size"]
#             )

#     # start vm
#     print(f"{vm.name} starting... ", end="")
#     print(wait_task_state(vm.PowerOn()))

#     # add to dns
#     if defaults["NETWORK"]["dns_api"] != "":
#         # assume technitium api url provided
#         # TODO: more integrations
#         result = technitium_add_record(hostname=vm_name, ip=vm_ip)
#         if isinstance(result, Response) and result.json()["status"] == "ok":
#             print(f"{vm.name} {vm_ip} added to dns")

#     if not wait_for_ssh(vm_ip):
#         fail(f"ssh failed for {vm_ip}")

#     # apply customisations to vm
#     print(f"{vm.name} configuring for {ez_node['product']}... ", end="")
#     if prepare_vm(
#         vm_name=vm.name,
#         vm_ip=vm_ip,
#         product_code=ez_node["product"],
#     ):
#         print(f"{vm_name} configured")

#     print("ok")

#     return vm.name, ez_node["product"]


# def create_vm(vm_folder: vim.Folder, vm_name, datastore, host, ez_node):
#     source_pool = host.parent.resourcePool
#     config = new_vm_spec(datastore=datastore, ez_node=ez_node, name=vm_name)
#     try:
#         task = vm_folder.CreateVm(config=config, pool=source_pool, host=host)
#         return wait_task_result(task=task)
#     except Exception as error:
#         fail(error)
#     return None


# def vms(vm_folder: vim.Folder):
#     return sorted(
#         [vm for vm in vm_folder.childEntity if isinstance(vm, vim.VirtualMachine)],
#         key=lambda x: x.name,
#     )


# def new_vm_spec(datastore, ez_node, name):
#     config = vim.vm.ConfigSpec()
#     config.annotation = f"Created by kayalab for {ez_node['name']} use"
#     config.memoryMB = int(ez_node["memGB"] * 1024)
#     config.guestId = "rhel8_64Guest"
#     config.name = name
#     config.numCPUs = ez_node["cores"]
#     config.cpuHotAddEnabled = True
#     config.memoryHotAddEnabled = True
#     files = vim.vm.FileInfo()
#     files.vmPathName = f"[{datastore.name}]"
#     config.files = files
#     return config


# def get_net_device(vm: vim.VirtualMachine):
#     nic_dev = None
#     for device in vm.config.hardware.device:
#         if isinstance(device, vim.vm.device.VirtualVmxnet3):
#             nic_dev = device

#     if nic_dev is None:
#         fail(f"Network device not found for {vm.config.name}")

#     return nic_dev


# def get_scsi_controller(vm: vim.VirtualMachine):
#     controller = None
#     for device in vm.config.hardware.device:
#         if isinstance(device, vim.vm.device.ParaVirtualSCSIController):
#             controller = device

#     if controller is None:
#         fail(f"PVSCSI Controller not found for {vm.config.name}")

#     return controller


# def add_new_disk_to_vm(
#     vm: vim.VirtualMachine, datastore: vim.Datastore, disk_size: int
# ):
#     """
#     Add new disk to vm
#     """
#     controller = get_scsi_controller(vm)

#     disk_sizeB = int(disk_size) * 1024 * 1024 * 1024
#     disk_number = len(controller.device)

#     disk_spec = vim.vm.device.VirtualDeviceSpec()
#     disk_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add
#     disk_spec.device = vim.vm.device.VirtualDisk()
#     disk_spec.device.backing = vim.vm.device.VirtualDisk.SeSparseBackingInfo()
#     disk_spec.device.backing.diskMode = "persistent"
#     disk_spec.device.backing.datastore = datastore
#     disk_spec.device.backing.fileName = (
#         f"[{datastore.name}] {vm.name}/{vm.name}-{disk_number}.vmdk"
#     )
#     disk_spec.device.unitNumber = disk_number
#     disk_spec.device.controllerKey = controller.key
#     disk_spec.device.key = -1
#     disk_spec.device.capacityInBytes = disk_sizeB
#     disk_spec.fileOperation = "create"

#     print(f"{vm.name} add {disk_size}GB disk...", end="")
#     print(reconfigure_vm(vm, disk_spec))

#     return vm


# def add_os_disk_to_vm(vm: vim.VirtualMachine, vmdk, disk_sizeGB: int):
#     controller = get_scsi_controller(vm)

#     disk_spec = vim.vm.device.VirtualDeviceSpec()
#     disk_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add
#     disk_spec.device = vim.vm.device.VirtualDisk()
#     disk_spec.device.deviceInfo = vim.Description(label="ez-os-template")
#     disk_spec.device.backing = vim.vm.device.VirtualDisk.FlatVer2BackingInfo()
#     disk_spec.device.backing.diskMode = (
#         vim.vm.device.VirtualDiskOption.DiskMode.persistent
#     )
#     # disk_spec.device.backing.thinProvisioned = True
#     # print(vmdk)
#     disk_spec.device.backing.fileName = f"{vmdk['folderPath']} {vmdk['file'].path}"

#     disk_spec.device.unitNumber = len(controller.device)
#     disk_spec.device.controllerKey = controller.key
#     # disk_spec.fileOperation = "create"

#     if reconfigure_vm(vm, disk_spec):
#         print(vmdk["file"].path)

#     # Resize OS disk
#     print("Resizing... ", end="")
#     disk_sizeB = int(disk_sizeGB) * 1024 * 1024 * 1024

#     vdisk = None
#     for device in vm.config.hardware.device:
#         if isinstance(device, vim.vm.device.VirtualDisk):
#             vdisk = device
#             break

#     disk_spec = vim.vm.device.VirtualDeviceSpec()
#     disk_spec.device = vdisk
#     disk_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.edit
#     disk_spec.device.capacityInBytes = disk_sizeB
#     if reconfigure_vm(vm, disk_spec):
#         print(f"{disk_sizeGB}GB")
#         return True

#     return False

# def confirm():
#     # dumb function
#     return True

# def delete_templates(si: vim.ServiceInstance):
#     content = si.RetrieveContent()
#     datacenter = select_from_objects(content.rootFolder.childEntity, "Data Center")

#     vmfolder = select_from_objects(datacenter.vmFolder.childEntity, "VM Folder")

#     vms = multiselect_from_list(
#         title="VMs",
#         list=[vm for vm in get_all_vms(vmfolder) if vm.config.template],
#         name_generator=lambda x: x.name,
#     )

#     if confirm(f"Are you sure to delete {[vm.name for vm in vms]}", default=False):
#         for vm in vms:
#             vm_name = vm.config.name
#             vm = typing.cast(vim.VirtualMachine, vm)
#             # disconnect template disk so it doesn't get deleted after last tempalte is deleted
#             print("Remove OS disk... ", end="")
#             if delete_os_disk_from_vm(vm):
#                 print(f"Destroy {vm.name}... ", end="")
#                 try:
#                     print(wait_task_state(vm.Destroy_Task()))
#                     yield vm_name
#                 except vim.fault.InvalidVmState as error:
#                     print(
#                         f"Failed to destroy {vm.name} with message: {error.faultMessage[0].message}"
#                     )

#     return None


# def delete_vms(si: vim.ServiceInstance):
#     content = si.RetrieveContent()
#     datacenter = select_from_objects(content.rootFolder.childEntity, "Data Center")

#     vmfolder = select_from_objects(datacenter.vmFolder.childEntity, "VM Folder")

#     vms = multiselect_from_list(
#         title="VMs",
#         list=[vm for vm in get_all_vms(vmfolder) if not vm.config.template],
#         name_generator=lambda x: x.name,
#     )

#     if len(vms) > 0 and confirm(
#         f"Are you sure to delete {[vm.name for vm in vms]}", default=False
#     ).ask(kbi_msg="Cancelled delete..."):
#         for vm in vms:
#             vm = typing.cast(vim.VirtualMachine, vm)
#             # capture name and ip for dns deletion
#             vm_name = vm.name
#             vm_ip = vm.summary.guest.ipAddress
#             if format(vm.runtime.powerState) == "poweredOn":
#                 print(f"Attempting to power off {vm.name}... ", end="")
#                 print(wait_task_state(vm.PowerOffVM_Task()))

#             print(f"Destroying {vm.name}... ", end="")
#             try:
#                 print(wait_task_state(vm.Destroy_Task()))
#                 if vm_ip:
#                     # remove from known_hosts
#                     subprocess.Popen(
#                         ["ssh-keygen", "-R", vm_ip],
#                         stdout=subprocess.DEVNULL,
#                         stderr=subprocess.DEVNULL,
#                     )
#                     subprocess.Popen(
#                         ["ssh-keygen", "-R", vm_name],
#                         stdout=subprocess.DEVNULL,
#                         stderr=subprocess.DEVNULL,
#                     )

#                     # remove from dns
#                     if config.get()["NETWORK"]["dns_api"] != "":
#                         result = technitium_delete_record(hostname=vm_name, ip=vm_ip)
#                         if (
#                             isinstance(result, Response)
#                             and result.json()["status"] == "ok"
#                         ):
#                             print(f"dns record deleted for {vm_name}")
#                         else:
#                             print(result.json())
#                 yield vm_name
#             except vim.fault.InvalidVmState as error:
#                 print(
#                     f"Failed to destroy {vm.name} with message: {error.faultMessage[0].message}"
#                 )

#     return None


# def delete_os_disk_from_vm(vm: vim.VirtualMachine):
#     os_disk = None
#     for dev in vm.config.hardware.device:
#         if (
#             isinstance(dev, vim.vm.device.VirtualDisk)
#             # and dev.deviceInfo.label == "ez-os-template"
#         ):
#             os_disk = dev
#     if os_disk is None:
#         fail(f"OS disk not found on {vm.name}")

#     # convert to vm first
#     # attach to the first host found for datastore
#     host: vim.HostSystem = os_disk.backing.datastore.host[0].key
#     try:
#         parent: vim.ComputeResource = host.parent
#         vm.MarkAsVirtualMachine(host=host, pool=parent.resourcePool)
#     except vmodl.fault.NotSupported:
#         print("ignoring...")
#         pass
#     disk_spec = vim.vm.device.VirtualDeviceSpec()
#     disk_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.remove
#     disk_spec.device = os_disk

#     if reconfigure_vm(vm, disk_spec):
#         print(os_disk.deviceInfo.label)
#         return True

#     return False


# def add_scsi_controller(vm):
#     controller_spec = vim.vm.device.VirtualDeviceSpec()
#     controller_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add
#     controller_spec.device = vim.vm.device.ParaVirtualSCSIController()
#     controller_spec.device.deviceInfo = vim.Description()
#     controller_spec.device.slotInfo = vim.vm.device.VirtualDevice.PciBusSlotInfo()
#     controller_spec.device.hotAddRemove = True
#     controller_spec.device.sharedBus = "noSharing"

#     return reconfigure_vm(vm, controller_spec)


# def add_nic_on_network(vm, network: vim.Network):
#     nic_spec = vim.vm.device.VirtualDeviceSpec()
#     nic_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add
#     nic_spec.device = vim.vm.device.VirtualVmxnet3()
#     nic_spec.device.deviceInfo = vim.Description()

#     if isinstance(network, vim.OpaqueNetwork):
#         nic_spec.device.backing = (
#             vim.vm.device.VirtualEthernetCard.OpaqueNetworkBackingInfo()
#         )
#         nic_spec.device.backing.opaqueNetworkType = network.summary.opaqueNetworkType
#         nic_spec.device.backing.opaqueNetworkId = network.summary.opaqueNetworkId
#     else:
#         nic_spec.device.backing = vim.vm.device.VirtualEthernetCard.NetworkBackingInfo()
#         nic_spec.device.backing.useAutoDetect = False
#         nic_spec.device.backing.deviceName = network.name

#     nic_spec.device.connectable = vim.vm.device.VirtualDevice.ConnectInfo()
#     nic_spec.device.connectable.startConnected = True
#     nic_spec.device.connectable.allowGuestControl = True
#     nic_spec.device.connectable.connected = False
#     nic_spec.device.connectable.status = "untried"
#     nic_spec.device.wakeOnLanEnabled = True
#     nic_spec.device.addressType = "generated"

#     return reconfigure_vm(vm, nic_spec)


# def reconfigure_vm(
#     vm: vim.VirtualMachine, device_spec: vim.vm.device.VirtualDeviceSpec
# ):
#     spec = vim.vm.ConfigSpec()
#     spec.deviceChange = [device_spec]
#     try:
#         return wait_task_state(vm.ReconfigVM_Task(spec=spec))
#     except Exception as error:
#         fail(error)


# def boot_from_cd(vm, iso_path):
#     cdspec = None
#     for device in vm.config.hardware.device:
#         if isinstance(device, vim.vm.device.VirtualCdrom):
#             cdspec = vim.vm.device.VirtualDeviceSpec()
#             cdspec.device = device
#             cdspec.operation = vim.vm.device.VirtualDeviceSpec.Operation.edit

#             cdspec.device.backing = vim.vm.device.VirtualCdrom.IsoBackingInfo()
#             for datastore in vm.datastore:
#                 cdspec.device.backing.datastore = datastore
#                 break
#             cdspec.device.backing.fileName = iso_path
#             cdspec.device.connectable = vim.vm.device.VirtualDevice.ConnectInfo()
#             cdspec.device.connectable.startConnected = True
#             cdspec.device.connectable.allowGuestControl = True

#     vmconf = vim.vm.ConfigSpec()
#     vmconf.deviceChange = [cdspec]
#     vmconf.bootOptions = vim.vm.BootOptions(
#         bootOrder=[vim.vm.BootOptions.BootableCdromDevice()]
#     )

#     task = vm.ReconfigVM_Task(vmconf)


# def get_userdata(
#     username,
#     password,
#     ssh_pubkey,
#     vm_name,
#     vm_domain,
# ):
#     return """
# #cloud-config
# hostname: {vm_name}
# fqdn: {vm_name}.{vm_domain}

# users:
#   - name: {username}
#     primary_group: {username}
#     passwd: {password}
#     shell: /bin/bash
#     sudo:  ALL=(ALL) NOPASSWD:ALL
#     groups: sudo, wheel
#     lock_passwd: false
#     ssh_authorized_keys:
#       - {ssh_pubkey}
# chpasswd:
#   expire: False

# package_upgrade: false
# """.format(
#         username=username,
#         password=password,
#         ssh_pubkey=ssh_pubkey,
#         vm_name=vm_name,
#         vm_domain=vm_domain,
#     )


# def get_metadata(
#     vm_name,
#     vm_ip,
#     vm_netmask,
#     vm_gateway,
#     vm_domain,
#     vm_nameserver,
# ):
#     return """instance-id: {vm_name}
# local-hostname: {vm_name}
# ssh_pwauth: true
# manage_resolv_conf: true
# network:
#   version: 1
#   config:
#     - type: physical
#       name: eth0
#       subnets:
#         - type: static
#           address: {vm_ip}
#           netmask: {vm_netmask}
#           gateway: {vm_gateway}
#     - type: nameserver
#       address:
#         - {vm_nameserver}
#       search:
#         - {vm_domain}
# """.format(
#         vm_name=vm_name,
#         vm_ip=vm_ip,
#         vm_netmask=vm_netmask,
#         vm_gateway=vm_gateway,
#         vm_domain=vm_domain,
#         vm_nameserver=vm_nameserver,
#     )



from ezlab.utils import *
from ezlab.parameters import *
from paramiko_expect import SSHClientInteraction


@fire_and_forget
def cluster_install(
    hosts: list[str],
    config: dict,
    queue: Queue,
    name: str = "ezlab",
    disks: str = "/dev/sdz",
    dryrun: bool = True,
):
    hostnames = []

    response = {}
    if dryrun:
        response["task"] = (
            f"Create Data Fabric cluster {name} on {','.join(hosts)} using disk(s) {disks}"
        )
        response["settings"] = config

    # lazy way to catch exceptions
    try:
        username = config.get("username")
        # used for connecting to nodes from this app
        keyfile = config.get("privatekeyfile")
        # used for connection between cluster nodes
        password = config.get("password")

        # get fqdn for each host
        if dryrun:

            def getsshcmd(host):
                return f"ssh {config.get('username')}@{host} 'hostname -f'"

            response["commands"] = [getsshcmd(host) for host in hosts]

        else:
            for host in hosts:
                print(f"Getting fqdn for {host}")
                for result in ssh_run_command(
                    host=host.strip(),
                    username=username,
                    keyfile=keyfile,
                    command="hostname -f",
                ):
                    hostnames.append(result)

        # addhosts = ([h for h in hosts if h["name"] != host["name"]],)
        # # to add to /etc/hosts
        # otherhosts = [
        #     f"{h['ip']} {h['name']}.{domain} {h['name']}" for h in addhosts
        # ].join("\n")

        commands: list[str] = INSTALL_COMMANDS[DF]

        if config.get("maprrepoislocal", False):
            # using local repository for mapr packages
            ezrepo: str = config.get("maprlocalrepo", "").rstrip("/")
            repouser = config["maprlocalrepousername"]
            repopass = config["maprlocalrepopassword"]

            wgetauthstring = (
                f"--user={repouser} --password={repopass}"
                if config.get("maprlocalrepoauth", False)
                else ""
            )

            proto, path, *_ = ezrepo.split("://")
            repourl = (
                f"{proto}://{repouser}:{repopass}@{path}"
                if config.get("maprlocalrepoauth", False)
                else ezrepo
            )

            commands.insert(
                0,
                f"[ -f /tmp/mapr-setup.sh ] || wget -q {wgetauthstring} {ezrepo}/installer/redhat/mapr-setup.sh -P /tmp; chmod +x /tmp/mapr-setup.sh",
            )

            commands.insert(
                1,
                f"curl -s -k https://127.0.0.1:9443/ > /dev/null || sudo /tmp/mapr-setup.sh -y -r {repourl}/",
            )

        else:
            # using HPE repository
            ezrepo = f"https://package.ezmeral.hpe.com/releases"
            repouser = config["maprrepouser"]
            repotoken = config["maprrepotoken"]

            commands.insert(
                0,
                f"[ -f /tmp/mapr-setup.sh ] || wget -q --user={repouser} --password= {repotoken} {ezrepo}/installer/redhat/mapr-setup.sh -P /tmp; chmod +x /tmp/mapr-setup.sh",
            )

            commands.insert(
                1,
                f"curl -s -k https://127.0.0.1:9443/ > /dev/null || sudo /tmp/mapr-setup.sh -y --repo-user {repouser} --repo-token {repotoken}",
            )

        first_host = hosts[0].strip()

        # Create and upload stanza file
        stanza = get_mapr_stanza(
            hosts=hostnames,
            username=username,
            password=password,
            disks=disks.split(","),
            cluster_name=name,
        )

        if dryrun:
            response["files"] = {"/tmp/mapr.stanza": stanza}

        else:
            print(f"{first_host} Writing /tmp/mapr.stanza")
            for result in ssh_content_into_file(
                host=first_host,
                username=username,
                keyfile=keyfile,
                content=stanza,
                filepath="/tmp/mapr.stanza",
            ):
                queue.put(f"{first_host}: {result}")

            queue.put(f"stanza copied to {first_host}")

            queue.put(f"starting installer on {first_host}")

        if dryrun:
            response["commands"].extend(commands)

        else:
            for command in commands:
                print(f"{first_host} Running: {command}")
                for result in ssh_run_command(
                    host=first_host,
                    username=username,
                    keyfile=keyfile,
                    command=command,
                ):
                    queue.put(f"{first_host} SSH: {result}")

            queue.put(
                f"echo Check installation process and results. Installer is available at https://{first_host}:9443/"
            )


    except Exception as error:
        print("Cluster creation exception", error)
        queue.put(TASK_FINISHED)
        if dryrun:
            return response
        else:
            return False

    queue.put(TASK_FINISHED)
    if dryrun:
        return response
    else:
        return True


@fire_and_forget
def client(
    server: str,
    client: str,
    username: str,
    password: str,
    repo: str,
    queue: Queue,
):
    repo_content = REPO_CONTENT.format(repo=repo)

    try:
        ssh_content_into_file(
            host=client,
            username=username,
            password=password,
            content=repo_content,
            filename="/etc/yum.repos.d/mapr.repo",
        )

        commands = INSTALL_COMMANDS[DFCLIENT]

        for file in DF_SECURE_FILES:
            commands.append(
                f"sudo curl -s --insecure --user mapr:mapr sftp://{server}{file} --output {file}"
            )

        # Run these last
        commands.append(
            f"sudo /opt/mapr/server/configure.sh -c -C {server} -N ezdf -secure"
        )
        commands.append("echo mapr | sudo -u mapr maprlogin password")

        for command in commands:
            # print(f"Runnning {command}")
            for result in ssh_run_command(
                host=client, username=username, password=password, command=command
            ):
                if result.strip() != "":
                    queue.put(f"{client} SSH: {result}")

    except Exception as error:
        print("Client configuration exception", error)
        queue.put(TASK_FINISHED)
        return False

    queue.put(TASK_FINISHED)
    return True


@fire_and_forget
def crosscluster(
    config: dict,
    settings: dict,
    queue: Queue,
    dryrun: bool = True,
):
    local = settings["crosslocalcldb"]
    remote = settings["crossremotecldb"]
    adminuser = settings["crossadminuser"]
    adminpassword = settings["crossadminpassword"]
    username = config["username"]
    keyfile = config["privatekeyfile"]

    queue.put(f"Create cross cluster connectivity between {local} and {remote}")

    response = {}
    if dryrun:
        response["task"] = f"Cross-cluster setup between {local} and {remote}"
        response["settings"] = {**config, **settings}

    else:  # get truststore passwords
        try:
            local_truststore_password = ""
            remote_truststore_password = ""

            for out in ssh_run_command(
                host=local,
                command="sudo grep ssl.server.truststore /opt/mapr/conf/store-passwords.txt | cut -d'=' -f2",
                username=username,
                keyfile=keyfile,
            ):
                local_truststore_password = out

            for out in ssh_run_command(
                host=remote,
                command="sudo grep ssl.server.truststore /opt/mapr/conf/store-passwords.txt | cut -d'=' -f2",
                username=username,
                keyfile=keyfile,
            ):
                remote_truststore_password = out

            if local_truststore_password and remote_truststore_password:
                queue.put(
                    f"using local truststore password: {local_truststore_password}"
                )
                queue.put(
                    f"using remote truststore password: {remote_truststore_password}"
                )
            else:
                queue.put(
                    f"ERROR: failed to get truststore passwords, local: {local_truststore_password}, remote: {remote_truststore_password}"
                )
                return False
        except Exception as error:
            print("Cross cluster setup exception", error)
            queue.put(TASK_FINISHED)
            return False

    cross_cluster_setup = f"sudo rm -rf /tmp/mapr-xcs; sudo -i -u {adminuser} /opt/mapr/server/configure-crosscluster.sh create all \
    -localcrossclusteruser {adminuser} -remotecrossclusteruser {adminuser} \
    -localtruststorepassword {local_truststore_password if local_truststore_password else 'LOCAL_TRUSTSTORE_PASSWORD'} \
    -remotetruststorepassword {remote_truststore_password if remote_truststore_password else 'REMOTE_TRUSTSTORE_PASSWORD'} \
    -remoteip {remote} -localuser {adminuser} -remoteuser {adminuser}"

    if dryrun:
        response["commands"] = (
            f"ssh {local} sudo grep ssl.server.truststore /opt/mapr/conf/store-passwords.txt | cut -d'=' -f2"
        )
        response["commands"] = (
            f"ssh {remote} sudo grep ssl.server.truststore /opt/mapr/conf/store-passwords.txt | cut -d'=' -f2"
        )
        response["commands"] = cross_cluster_setup
        queue.put(TASK_FINISHED)
        return response

    else:
        try:
            queue.put(
                "If fails run the configure-crosscluster.sh command on one of the cluster nodes:"
            )
            queue.put(cross_cluster_setup)

            client = SSHClient()
            client.set_missing_host_key_policy(AutoAddPolicy)

            client.connect(hostname=local, username=username, key_filename=keyfile)

            interact = SSHClientInteraction(client, timeout=300, display=True)
            prompt = f".*[{username}@{local} ~]$.*"
            interact.expect(prompt)
            # print(interact.current_output_clean)

            interact.send(cross_cluster_setup)

            interact.expect("Enter password for mapr user.*")
            interact.send(adminpassword)
            interact.expect("Enter password for mapr user.*")
            interact.send(adminpassword)

            interact.expect(prompt)
            interact.send("exit")
            interact.expect()
        except Exception as error:
            print("Cross cluster setup exception", error)
            queue.put(TASK_FINISHED)
            return False

        queue.put(TASK_FINISHED)
        return True


def get_mapr_stanza(
    hosts: list, username: str, password: str, disks: list, cluster_name: str
):
    """
    Return stanza for Data Fabric installation
    """
    hosts_yaml = ""
    for host in hosts:
        hosts_yaml += f"  -   {host}\n"

    disks_yaml = ""
    for disk in disks:
        disks_yaml += f"  -   {disk}\n"

    return f"""
environment:
  mapr_core_version: 7.6.0
config:
  hosts:
{hosts_yaml}
  ssh_id: {username}
  ssh_password: {password}
  enable_nfs: False
  db_admin_user: root
  db_admin_password: mapr123
  log_admin_password: mapr123
  metrics_ui_admin_password: mapr123
  enable_encryption_at_rest: True
  license_type: M7
  mep_version: 9.2.1
  disks:
{disks_yaml}
  disk_format: true
  disk_stripe: 1
  cluster_name: {cluster_name}
  services:
    template-05-converged:
    # template-30-maprdb:
    # template-60-maprxd:
    mapr-keycloak:
      enabled: True
    mapr-hivemetastore:
      database:
        name: hive
        user: hive
        password: mapr123
    mapr-grafana:
      enabled: False
    mapr-opentsdb:
      enabled: False
    mapr-collectd:
      enabled: False
    mapr-fluentd:
      enabled: False
    mapr-kibana:
      enabled: False
    mapr-elasticsearch:
      enabled: False
    mapr-data-access-gateway:
    mapr-mastgateway:

"""



import base64
from enum import Enum
from io import BytesIO
from ipaddress import IPv4Address, ip_address, ip_network
from socket import gethostbyaddr
import socket
from uuid import uuid4

from urllib import parse
import os
import shlex
import subprocess
from threading import Thread
from time import sleep
from paramiko import (
    AuthenticationException,
    AutoAddPolicy,
    BadHostKeyException,
    SSHException,
    client as pclient,
)
from questionary import Choice, checkbox, prompt, select
from rich import print

from .parameters import ez_nodes, ez_files, ez_commands
from . import config

defaults = config.get()
if len(defaults.keys()) == 0:
    config.set()
    defaults = config.get()
vm_network = defaults["NETWORK"]["vm_cidr"]
no_proxy = vm_network
# no_proxy = ",".join([str(h) for h in ip_network(vm_network).hosts()])
vm_proxy = defaults["NETWORK"]["proxy"] if "proxy" in defaults["NETWORK"] else ""
if "yum_proxy" in defaults["NETWORK"]:
    no_proxy += f",{parse.urlsplit(defaults['NETWORK']['yum_proxy']).hostname}"
vm_domain = defaults["NETWORK"]["vm_domain"]


class target_names(str, Enum):
    """
    Supported target platforms
    """

    pve = "pve"
    vmw = "vmw"


class target_classes(str, Enum):
    """
    API classes for target platforms
    """

    pve = "ProxmoxAPI"
    vmw = "vim.ServiceInstance"


def fail(message):
    """
    Print the message and gracefully exit
    Params:
        <message>: parameter to print
    """
    print("Operation failed!")
    print(message)
    exit(1)


def select_from_list(
    title: str, list: list, name_generator=lambda n: n, value_generator=lambda v: v
):
    """
    Select an item from a list of objects.
    Params:
        <title>           : Title to show for selection
        <list>            : List of objects (or a simple list(str))
        <name_generator>  : Optional function to get display name for the selection, defaults to the object (if object has __str__ or __name__ method it is returned by default)
        <value_generator> : Optional function to get value for the selection, defaults to the object

    Returns: selected object | <value>
    """
    if not list or len(list) == 0:
        print(f"No {title} found")
        return None
    # generate choices from list with provided name/value field transformers
    choices = [make_choice(name_generator(x), value_generator(x)) for x in list]
    choice = select(title, choices=choices).ask()
    if not choice:
        fail(f"Nothing selected for {title}")
        return None
    # find and return the object from list matching the choice
    return [x for x in list if choice == value_generator(x)][0]


def multiselect_from_list(
    title: str, list: list, name_generator=lambda n: n, value_generator=lambda v: v
):
    """
    Select multiple items from a list of objects.
        Params:
        <title>           : Title to show for selection
        <list>            : List of objects (or a simple list(str))
        <name_generator>  : Optional function to get display name for the selection, defaults to the object (if object has __str__ or __name__ method it is returned by default)
        <value_generator> : Optional function to get value for the selection, defaults to the object

    Returns: list of selected object | <value>

    """
    if not list or len(list) == 0:
        print(f"No {title} found")
        return []
    choices = [make_choice(name_generator(x), value_generator(x)) for x in list]
    choice = checkbox(title, choices=choices).ask()
    if not choice:
        fail(f"Nothing selected for {title}")
    return [x for x in list if value_generator(x) in choice]


def make_choice(name, value=None) -> Choice:
    if not value:
        value = name
    return Choice.build({"name": name, "value": value})


def ask_customisations(**kwargs):
    """
    Ask user for customised values
    Returns <dict>: with selection
    """
    # selecting 3rd available ip as the first vm ip
    recommended_first_ip = str(
        list(ip_network(defaults["NETWORK"]["vm_cidr"]).hosts())[3]
    )
    customisation = [
        {
            "type": "text",
            "name": "vmname",
            "message": "First VM name",
            "default": "ezvm",
        },
        {
            "type": "text",
            "name": "first_vm_ip",
            "message": "First VM IP or 'dhcp'",
            "default": recommended_first_ip,
        },
    ]
    return prompt(customisation, **kwargs)


def get_ssh_key_file():
    return os.path.abspath(os.path.expanduser(defaults["CLOUDINIT"]["ssh_keyfile"]))


def get_customisations():
    """
    Ask user for customisations and verify if they are valid

    """

    vm_ip = ""
    customisations = ask_customisations()
    try:
        vm_ip: str = customisations["first_vm_ip"] or ""

    except KeyError:
        fail("missing information")

    if vm_ip != "dhcp":
        try:
            # check if ip settings are fitting
            if ip_address(vm_ip) not in ip_network(vm_network):
                fail(f"{vm_ip} is not in configured network {vm_network}")
            # confirm enough IP addresses are available in the chosen block for selected node counts
        except ValueError as error:
            print(error)
            fail(f"Invalid IP address or network: {vm_ip} in {vm_network}")

    return customisations


def select_nodes():
    selected = checkbox("Select nodes to deploy", choices=ez_nodes).ask() or []
    nodes = []
    for node in selected:
        nodes.extend([n for n in ez_nodes if n["name"] == node])
    return nodes


def prepare_vm(vm_name: str, vm_ip: str, product_code: str = ""):
    """
    Configure VMs for selected product, taking care of pre-requisites
    Params:
        <vm_name>      : VM name to use in hosts and various system files
        <vm_ip>        : VM IP address to use in hosts and various system files
        <product_code> : Prepare for ezdf | ezua | generic installations

    Returns <bool>     : True in succesful completion, else False
    """

    # files to write
    add_files: dict = ez_files[product_code]

    # update /etc/hosts
    add_files.update(
        {
            "/etc/hosts": f"127.0.0.1 localhost.localdomain localhost\n{vm_ip} {vm_name}.{vm_domain} {vm_name}\n"
        }
    )

    if vm_proxy != "":
        add_files = update_proxy_files(add_files)

    # TODO: Not tested
    if (
        "airgap_registry" in defaults["NETWORK"]
        and defaults["NETWORK"]["airgap_registry"].split("://")[0] == "http"
    ):
        add_files.update({"/etc/docker/daemon.json": get_insecure_registry()})

    if "yum_proxy" in defaults["NETWORK"]:
        repo_content = get_repo_content()
        add_files.update({"/etc/yum.repos.d/nexus.repo": repo_content})

    for file, content in add_files.items():
        # print(f"Writing: {file}")
        if not sftp_write_into_file(
            host=vm_ip,
            username=defaults["CLOUDINIT"]["vm_username"],
            filename=file,
            content=content,
        ):
            return False

    # commands to execute
    commands = list(ez_commands[product_code])

    if "yum_proxy" in defaults["NETWORK"]:
        commands.insert(
            2, "sudo dnf config-manager --set-disabled baseos appstream extras"
        )

    elif vm_proxy != "":
        commands.insert(
            1,
            f"echo proxy={vm_proxy} | sudo tee -a /etc/dnf/dnf.conf > /dev/null",
        )

    for command in commands:
        # print(f"Running: {command}")
        for output in ssh_run_command(vm_ip, command=command):
            print(output)

    return True


def technitium_add_record(hostname: str, ip: str):
    """
    Add DNS records for hostname in Technitium DNS
    Params:
        <hostname>       : hostname to add in DNS
        <ip>             : IP Address to add in DNS
        # <admin_user>     : Technitium admin user
        # <admin_password> : Technitium admin password, defaults to vm_password provided during install/config

    Returns result from API call and prints the error if any
    """
    from tdnss.connection import Connection

    c = Connection(defaults["NETWORK"]["dns_api"])
    res = c.login(
        username=defaults["CLOUDINIT"]["vm_username"],
        password=defaults["CLOUDINIT"]["vm_password"],
    )
    result = None
    if res.status == 1:
        print(f"Cannot add DNS record! Add {ip}: {hostname} to your DNS!")
        print(res.message)
    else:
        result = c._get(
            "zones/records/add",
            {
                "domain": f"{hostname}.{vm_domain}",
                "zone": vm_domain,
                "type": "A",
                "overwrite": "true",
                "ipAddress": ip,
                "ptr": "true",
            },
        )
    c.logout()
    return result


# def get_dns_admin_password():
#     admin_password = os.getenv("EZLAB_DNSPASS")
#     if admin_password is None:
#         admin_password = password("DNS Admin password").ask()

#     return admin_password


def technitium_delete_record(hostname: str, ip: str):
    """
    Delete DNS records for hostname in Technitium DNS
    Params:
        <hostname>       : hostname to delete in DNS
        <ip>             : IP Address to delete in DNS
        # <admin_user>     : Technitium admin user
        # <admin_password> : Technitium admin password, defaults to vm_password provided during install/config

    Returns result from API call and prints the error if any
    """

    from tdnss.connection import Connection

    c = Connection(defaults["NETWORK"]["dns_api"])
    res = c.login(
        username=defaults["CLOUDINIT"]["vm_username"],
        password=defaults["CLOUDINIT"]["vm_password"],
    )
    if res.status == 1:
        print(f"Cannot delete DNS record! Remove {ip}: {hostname} on your DNS!!!")
        print(res.message)
    else:
        res = c._get(
            "zones/records/delete",
            {
                "domain": f"{hostname}.{vm_domain}",
                "zone": vm_domain,
                "type": "A",
                "ipAddress": ip,
                "overwrite": "true",
                "ptr": "true",
            },
        )
    c.logout()
    return res


def ssh_run_command(
    host: str, command: str, username: str = defaults["CLOUDINIT"]["vm_username"], password: str = ""
):
    """
    Run commands over ssh using default credentials
    Params:
        <host>              : hostname/IP to SSH
        <command>           : command to run
        <username>          : (optional) username to authenticate
        <password>          : (optional) password for <username>
    Returns:
        Generator with output. Calling function should process output with a for loop.
        Prints out stderr directly into terminal.

    """
    client = pclient.SSHClient()
    client.set_missing_host_key_policy(AutoAddPolicy)
    command_output = None
    try:
        if password == "":
            client.connect(
                host,
                username=username,
                key_filename=get_ssh_key_file(),
                timeout=60,
            )
        else:
            client.connect(
                host,
                username=username,
                password=password,
                timeout=60,
            )

        _stdin, _stdout, _stderr = client.exec_command(command, get_pty=True)
        command_errors = _stderr.read().decode().strip()
        command_output = _stdout.read().decode().strip()
        if command_errors and command_errors != "":
            print(f"STDERR: {command_errors}")
        if command_output and command_output != "":
            yield command_output
    except Exception as error:
        fail(error)
    finally:
        client.close()

    # return command_output


def wait_for_ssh(host):
    client = pclient.SSHClient()
    client.set_missing_host_key_policy(AutoAddPolicy)
    try:
        while True:
            client.connect(
                host,
                username=defaults["CLOUDINIT"]["vm_username"],
                key_filename=get_ssh_key_file(),
            )
            break
    except (
        BadHostKeyException,
        AuthenticationException,
        SSHException,
        socket.error,
    ):
        print(f"{host} wait for ssh...")
        sleep(5)
    finally:
        client.close()

    return True


def sftp_write_into_file(host, username, filename, content):
    """
    Create files over SFTP using default credentials, putting the content into the file.
    Note:
        The file will be created in temp location as the configured user, but then will be moved to location with sudo
    Params:
        <host>       : hostname/IP to SSH
        <username>   : connect as username
        <filename>   : full path to file in remote system
        <content>    : file content

    """

    client = pclient.SSHClient()
    # client.set_missing_host_key_policy(WarningPolicy)
    client.set_missing_host_key_policy(AutoAddPolicy)
    ftp = None
    try:
        client.connect(host, username=username, key_filename=get_ssh_key_file())
        ftp = client.open_sftp()

        seq = str(uuid4().hex)

        ftp.putfo(BytesIO(content.encode()), f"/tmp/{seq}")
        client.exec_command(
            f"sudo mkdir -p $(dirname {filename}) && sudo mv /tmp/{seq} {filename} && chown root:root {filename} && chmod 644 {filename}"
        )
    except (
        BadHostKeyException,
        AuthenticationException,
        SSHException,
        socket.error,
    ) as e:
        print(e)
        sleep(5)
    finally:
        if ftp:
            ftp.close()
        client.close()

    return True


class ReturningThread(Thread):
    """Create a Thread which returns a value

    Args:
        Thread (Thread): Thread to run and return the result/return value
    """

    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        Thread.__init__(self, group, target, name, args, kwargs)
        self._return = None

    def run(self):
        if self._target is not None:  # type: ignore
            self._return = self._target(*self._args, **self._kwargs)  # type: ignore

    def join(self, *args):
        Thread.join(self, *args)
        return self._return


def start_tasks_delayed(threads: list, delay: int = 0):
    """Start a thread for each task with an optional delay

    Args:
        threads (list<thread>): list of threads to start
        delay (int): delay in seconds
    """
    for t in threads:
        t.start()
        sleep(delay)  # wait few sec to give api time against race condition


def wait_tasks(threads):
    """wait for finish and return result from threads as they come"""
    for t in threads:
        yield t.join()


def toB64(string: str):
    return base64.b64encode(string.encode("ascii")).decode("utf-8")


def execute(cmd: str):
    # sub_env = os.environ.copy()
    # for line in get_proxy_environment().splitlines():
    #     if line.strip() != "":
    #         key, value = line.split("=")
    #         sub_env[key] = value
    proc = subprocess.Popen(
        args=shlex.split(cmd),
        stdout=subprocess.PIPE,
        universal_newlines=True,
        # env=sub_env,
    )

    while proc.stdout:
        output = proc.stdout.readline().strip()
        if output == "" and proc.poll() is not None:
            break
        if output:
            yield f">>> {output}"
    rc = proc.poll()

    # if return_code:
    #     print(f"Error running command: {cmd}")
    #     return "Error"
    # raise subprocess.CalledProcessError(return_code, cmd)


def find_app(name: str):
    """Return executable path for `name` on PATH or in CWD."""
    from shutil import which

    return which(cmd=name, path=f"{os.environ.get('PATH')}:{os.getcwd()}")


def get_proxy_environment():
    # TODO: refactor to return dict and get_proxy_profile() and /etc/environment files can be generated with that
    return f"""
HTTP_PROXY={vm_proxy}
http_proxy={vm_proxy}
HTTPS_PROXY={vm_proxy}
https_proxy={vm_proxy}
NO_PROXY=10.96.0.0/12,10.224.0.0/16,10.43.0.0/16,.external.hpe.local,localhost,.cluster.local,.svc,.default.svc,127.0.0.1,169.254.169.254,.{vm_domain},{no_proxy}
no_proxy=10.96.0.0/12,10.224.0.0/16,10.43.0.0/16,.external.hpe.local,localhost,.cluster.local,.svc,.default.svc,127.0.0.1,169.254.169.254,.{vm_domain},{no_proxy}

"""


def get_proxy_profile():
    return f"""
export HTTP_PROXY={vm_proxy}
export http_proxy={vm_proxy}
export HTTPS_PROXY={vm_proxy}
export https_proxy={vm_proxy}
export NO_PROXY=10.96.0.0/12,10.224.0.0/16,10.43.0.0/16,.external.hpe.local,localhost,.cluster.local,.svc,.default.svc,127.0.0.1,169.254.169.254,.{vm_domain},{no_proxy}
export no_proxy=10.96.0.0/12,10.224.0.0/16,10.43.0.0/16,.external.hpe.local,localhost,.cluster.local,.svc,.default.svc,127.0.0.1,169.254.169.254,.{vm_domain},{no_proxy}

"""


def get_repo_content():
    repo_content = ""
    for repo in ["BaseOS", "AppStream", "PowerTools", "extras"]:
        repo_content += f"""
[nexus-{repo.lower()}]
name = Nexus {repo}
enabled = 1
gpgcheck = 0
baseurl = {defaults["NETWORK"]["yum_proxy"]}/$releasever/{repo}/$basearch/os
ui_repoid_vars = releasever basearch
priority=1
proxy=

"""
    return repo_content


def update_proxy_files(file_content_object: dict):
    file_content_object.update({"/etc/environment": get_proxy_environment()})
    file_content_object.update({"/etc/profile.d/proxy.sh": get_proxy_profile()})
    # file_content_object.update(
    #     {f"/home/{defaults['CLOUDINIT']['vm_username']}/.bashrc": get_proxy_profile()}
    # )
    containerd_proxy = f"""
[Service]
EnvironmentFile=/etc/environment
"""
    # [Service]
    # Environment="HTTP_PROXY={vm_proxy}"
    # Environment="HTTPS_PROXY={vm_proxy}"
    # Environment="NO_PROXY=localhost,127.0.0.1,.{vm_domain},{no_proxy}"
    file_content_object.update(
        {"/etc/systemd/system/containerd.service.d/http-proxy.conf": containerd_proxy}
    )

    return file_content_object


def get_insecure_registry():
    return f"""
{
  "log-driver": "journald",
  "insecure-registries" : [ {defaults['NETWORK']['airgap_registry']} ]
}
"""


def validate_ip(ip_string: str) -> bool:
    try:
        return isinstance(ip_address(ip_string), IPv4Address)
    except:
        return False


def get_fqdn(ip_string: str) -> str:
    fqdn, _, _ = gethostbyaddr(ip_string)
    return fqdn



from typing import Annotated, List
from urllib.parse import quote
from paramiko import AutoAddPolicy, SSHClient
from paramiko_expect import SSHClientInteraction

import typer

from . import config

from .common import (
    fail,
    sftp_write_into_file,
    ssh_run_command,
)

app = typer.Typer(add_completion=True, no_args_is_help=True)

defaults = config.get()


@app.command(
    help="Create Data Fabric cluster with provided hosts",
    short_help="Create DF cluster",
    no_args_is_help=True,
)
def cluster(
    hosts: List[str],
    name: Annotated[str, typer.Option("--name", "-n", help="cluster name")] = "ezlab",
    repouser: Annotated[str, typer.Option("--username", "-u", help="repo user")] = "",
    repotoken: Annotated[
        str,
        typer.Option(
            "--token",
            "-t",
            hide_input=True,
            help="password or token to authenticate for the repo",
        ),
    ] = "",
):
    hostnames = []
    for host in hosts:
        for result in ssh_run_command(
            host=f"{host}.{defaults['NETWORK']['vm_domain']}", command="hostname -f"
        ):
            hostnames.append(result)

    first_host = hosts[0]

    print(first_host)

    ez_repo = (
        defaults["NETWORK"]["mapr_repo"]
        if "mapr_repo" in defaults["NETWORK"]
        else f"https://{repouser}:{repotoken}@package.ezmeral.hpe.com/releases"
    )

    sftp_write_into_file(
        host=first_host,
        username=defaults["CLOUDINIT"]["vm_username"],
        filename="/tmp/mapr.stanza",
        content=get_mapr_stanza(
            hosts=hostnames,
            username=defaults["CLOUDINIT"]["vm_username"],
            password=defaults["CLOUDINIT"]["vm_password"],
            disks=["/dev/sdb"],
            cluster_name=name,
        ),
    )
    print(f"stanza copied to {first_host}")

    # for wget string
    user_str = quote(f"--user={repouser}" if repouser else "")
    token_str = f"--password={repotoken}" if repotoken else ""

    commands = [
        f"[ -f /tmp/mapr-setup.sh ] || wget -q {user_str} {token_str} {ez_repo}/installer/redhat/mapr-setup.sh -P /tmp",
        "chmod +x /tmp/mapr-setup.sh",
        f"curl -s -k https://127.0.0.1:9443/ > /dev/null || sudo /tmp/mapr-setup.sh -y -r {ez_repo}",
        "[ -f /opt/mapr/installer/bin/mapr-installer-cli ] && echo Starting installation, this may take ~30 minutes... && sleep 30",
        "echo Installer at: https://$(hostname -f):9443/",
        "echo -n 'Starting at: '; date",
        # "[ -f /opt/mapr/installer/bin/mapr-installer-cli ] && echo Monitor progres on Installer web, do not use navigation links!",
        "echo y | sudo /opt/mapr/installer/bin/mapr-installer-cli install -nvp -t /tmp/mapr.stanza -u mapr:mapr@127.0.0.1:9443",
        "echo -n 'Finished at: '; date",
    ]

    print(f"starting installer on {first_host}")
    for command in commands:
        for result in ssh_run_command(host=first_host, command=command):
            print(result)

    print(
        f"echo Installation complete. Install license file and configure cluster on https://{first_host}:8443/"
    )


@app.command(
    help="Configure DF client for specified cluster (CLDB)",
    short_help="Create a DF client",
    no_args_is_help=True,
)
def dfclient(
    server: Annotated[str, typer.Option("--server", "-s")],
    client: Annotated[str, typer.Option("--client", "-c")],
    user: Annotated[str, typer.Option("--username", "-u", help="repo user")] = "",
    token: Annotated[
        str,
        typer.Option(
            "--token",
            "-t",
            hide_input=True,
        ),
    ] = "",
):
    ez_repo = (
        defaults["NETWORK"]["mapr_repo"]
        if "mapr_repo" in defaults["NETWORK"]
        else f"https://{user}:{token}@package.ezmeral.hpe.com/releases"
    )

    repo_content = """
[ezmeral]
name = Ezmeral Packages
enabled = 1
gpgcheck = 0
baseurl = {ez_repo}/v7.5.0/redhat
priority=1
proxy=

[ezmeral-eep]
name = Ezmeral EEP Packages
enabled = 1
gpgcheck = 0
baseurl = {ez_repo}/MEP/MEP-9.2.0/redhat
priority=1
proxy=

""".format(
        ez_repo=ez_repo
    )

    sftp_write_into_file(
        host=client,
        username=defaults["CLOUDINIT"]["vm_username"],
        filename="/etc/yum.repos.d/mapr.repo",
        content=repo_content,
    )

    commands = [
        "sudo dnf update -y -q",
        "sudo useradd -md /home/mapr -u 5000 -U -s /bin/bash mapr",
        "echo mapr:mapr | sudo chpasswd",
        "sudo dnf install -y -q mapr-client java-11-openjdk",
    ]
    files = [
        "/opt/mapr/conf/ssl_truststore",
        "/opt/mapr/conf/ssl_truststore.p12",
        "/opt/mapr/conf/ssl_truststore.pem",
        "/opt/mapr/conf/maprtrustcreds.conf",
        "/opt/mapr/conf/maprtrustcreds.jceks",
        "/opt/mapr/conf/ssl_keystore-signed.pem",
    ]
    for file in files:
        commands.append(
            f"sudo curl -s --insecure --user mapr:mapr sftp://{server}{file} --output {file}"
        )

    # Run these last
    commands.append(
        f"sudo /opt/mapr/server/configure.sh -c -C {server} -N ezdf -secure"
    )
    commands.append("echo mapr | sudo -u mapr maprlogin password")

    for command in commands:
        print(f"Runnning {command}")
        for result in ssh_run_command(host=client, command=command):
            print(result)


@app.command(
    help="Create cross-cluster connectivity between local and remote",
    short_help="Configure cross-cluster",
    no_args_is_help=True,
)
def cross(
    local: Annotated[str, typer.Option("--local", "-l")],
    remote: Annotated[str, typer.Option("--remote", "-r")],
    luser: Annotated[
        str, typer.Option("--luser", help="local cluster admin user")
    ] = "mapr",
    ruser: Annotated[
        str, typer.Option("--ruser", help="local cluster admin user")
    ] = "mapr",
    password: Annotated[
        str, typer.Option("--password", "-p", help="Admin user (mapr) password")
    ] = "mapr",
):
    print(f"Create cross cluster connectivity between {local} and {remote}")

    local_truststore_password = ""
    remote_truststore_password = ""
    for out in ssh_run_command(
        local,
        "grep ssl.server.truststore /opt/mapr/conf/store-passwords.txt | cut -d'=' -f2",
        username=luser,
        password=password,
    ):
        local_truststore_password = out

    for out in ssh_run_command(
        remote,
        "grep ssl.server.truststore /opt/mapr/conf/store-passwords.txt | cut -d'=' -f2",
        username=ruser,
        password=password,
    ):
        remote_truststore_password = out

    if local_truststore_password and remote_truststore_password:
        print(f"using local truststore password: {local_truststore_password}")
        print(f"using remote truststore password: {remote_truststore_password}")
    else:
        fail("failed to get truststore passwords")

    cross_cluster_setup = f"rm -rf /tmp/mapr-xcs; /opt/mapr/server/configure-crosscluster.sh create all \
    -localcrossclusteruser {luser} -remotecrossclusteruser {ruser} \
    -localtruststorepassword {local_truststore_password} -remotetruststorepassword {remote_truststore_password} \
    -remoteip {remote} -localuser {luser} -remoteuser {ruser} "

    # for out in ssh_run_command(host=local, command=cross_cluster_setup, username=luser, password=password):
    #     print(out)
    client = SSHClient()
    client.set_missing_host_key_policy(AutoAddPolicy)

    client.connect(hostname=local, username=luser, password=password)

    interact = SSHClientInteraction(client, timeout=300, display=True)
    prompt = f".*[{luser}@{local} ~]$.*"
    interact.expect(prompt)
    # print(interact.current_output_clean)

    interact.send(cross_cluster_setup)

    interact.expect("Enter password for mapr user.*")
    interact.send(password)
    interact.expect("Enter password for mapr user.*")
    interact.send(password)

    interact.expect(prompt)
    interact.send("exit")
    interact.expect()


def get_mapr_stanza(
    hosts: list, username: str, password: str, disks: list, cluster_name: str
):
    """
    Return stanza for Data Fabric installation
    """
    hosts_yaml = ""
    for host in hosts:
        hosts_yaml += f"  -   {host}\n"

    disks_yaml = ""
    for disk in disks:
        disks_yaml += f"  -   {disk}\n"

    return f"""
environment:
  mapr_core_version: 7.5.0
config:
  hosts:
{hosts_yaml}
  ssh_id: {username}
  ssh_password: {password}
  enable_nfs: False
  db_admin_user: root
  db_admin_password: mapr123
  log_admin_password: mapr123
  metrics_ui_admin_password: mapr123
  enable_encryption_at_rest: True
  license_type: M7
  mep_version: 9.2.0
  disks:
{disks_yaml}
  disk_format: true
  disk_stripe: 1
  cluster_name: {cluster_name}
  services:
    # template-05-converged:
    template-30-maprdb3:
    mapr-hivemetastore:
      database:
        name: hive
        user: hive
        password: mapr123
    mapr-grafana:
      enabled: True
    mapr-opentsdb:
      enabled: True
    mapr-collectd:
    mapr-fluentd:
    mapr-kibana:
      enabled: False
    mapr-elasticsearch:
      enabled: Flase
    mapr-data-access-gateway:
    mapr-mastgateway:

"""
